import numpy as np
from .utils import _parse_to_list, _concatenate_to_array
from .core_objects import Signal, Module, Network
from typing import List, Iterable, Union, Callable


# flake8: noqa: C901
def finite_difference(blk: Module, fromsig: Union[Signal, Iterable[Signal]] = None,
                      tosig: Union[Signal, Iterable[Signal]] = None,
                      dx: float = 1e-8, tol: float = 1e-5, random: bool = True, use_df: list = None,
                      test_fn: Callable = None, keep_zero_structure=True, verbose=True):
    """ Performs a finite difference check on the given Module or Network

    Args:
        blk: The module or network
        fromsig (optional): Specify input signals of interest
        tosig (optional): Specify output signals of interest

    Keyword Args:
        dx: Perturbation size
        tol: Tolerance
        random: Randomize sensitivity data
        use_df: Give pre-defined sensitivity data
        test_fn: A generic test function (x, dx, df_an, df_fd)
        keep_zero_structure: If ``True`` variables that are ``0`` are not perturbed
        verbose: Print extra information to console
    """
    print("=========================================================================================================\n")
    print("Starting finite difference of \"{0}\" with dx = {1}, and tol = {2}".format(type(blk).__name__, dx, tol))

    # Parse inputs and outputs to be finite-differenced
    inps = blk.sig_in if fromsig is None else _parse_to_list(fromsig)
    outps = blk.sig_out if tosig is None else _parse_to_list(tosig)

    # In case a Network is passed, only the blocks connecting input and output need execution
    if isinstance(blk, Network):
        i_first, i_last = -1, -1
        for i, b in enumerate(blk.mods):
            # Find the first module which requires any of the input signals
            if i_first < 0 and any([s in inps for s in b.sig_in]):
                i_first = i
            # Find the last module that generates any of the output signals
            if any([s in outps for s in b.sig_out]):
                i_last = i
        blks_pre = Network(blk.mods[:i_first])
        if i_last < 0:
            blk = Network(blk.mods[i_first:])
        else:
            blk = Network(blk.mods[i_first:i_last+1])
        # Precompute only once for any blocks before first occurrence of <inps>
        blks_pre.response()

    print("Inputs:")
    if verbose:
        [print("{}\t{} = {}".format(i, s.tag, s.state)) for i, s in enumerate(inps)]
    else:
        print(", ".join([s.tag for s in inps]))
    print("")

    # Setup some internal storage
    f0 = [None for _ in outps]  # Response values at original inputs
    df_an = [np.empty(0) for _ in outps]  # Analytical output sensitivities
    dx_an = [[np.empty(0) for _ in inps] for _ in outps]

    # Initial reset in case some memory is still left
    blk.reset()

    # Perform response
    blk.response()

    print("Outputs:")
    if not verbose:
        print(", ".join([s.tag for s in outps]))

    # Get analytical response and sensitivities, by looping over all outputs
    for Iout, Sout in enumerate(outps):
        # Obtain the output state
        output = Sout.state
        if verbose:
            print("{}\t{} = {}".format(Iout, Sout.tag, output))

        # Store the output value
        f0[Iout] = (output.copy() if hasattr(output, "copy") else output)

        # Get the output state shape
        shape = (output.shape if hasattr(output, "shape") else ())

        # Generate a (random) sensitivity for output signal
        if use_df is not None:
            df_an[Iout] = use_df[Iout]
        else:
            if random:
                df_an[Iout] = np.random.rand(*shape)
            else:
                df_an[Iout] = np.ones(shape)

            if np.iscomplexobj(output):
                if random:
                    df_an[Iout] = df_an[Iout] + 1j * np.random.rand(*shape)
                else:
                    df_an[Iout] = df_an[Iout] + 1j * np.ones(shape)

        # Set the output sensitivity
        Sout.sensitivity = df_an[Iout]

        # Perform the analytical sensitivity calculation
        blk.sensitivity()

        # Store all input sensitivities for this output
        for Iin, Sin in enumerate(inps):
            sens = Sin.sensitivity
            dx_an[Iout][Iin] = (sens.copy() if hasattr(sens, "copy") else sens)

        # Reset the sensitivities for next output
        blk.reset()

    # Perturb each of the input signals
    for Iin, Sin in enumerate(inps):
        print("___________________________________________________")
        print("Perturbing input {} \"{}\"...\n".format(Iin, Sin.tag))

        # Get input state
        x = Sin.state

        try:
            # Get iterator for x
            it = np.nditer(x, flags=['c_index', 'multi_index'], op_flags=['readwrite'])
            is_iterable = True
        except TypeError:
            it = np.nditer(np.array(x), flags=['c_index', 'multi_index'], op_flags=['readwrite'])
            is_iterable = False

        i_failed, i_tested = 0, 0
        # Loop over all values in x
        while not it.finished:
            # Get original value and do the perturbation
            if is_iterable:
                x0 = it[0].copy()
                if x0 == 0 and keep_zero_structure:
                    it.iternext()
                    continue
                it[0] += dx
                Sin.state = x
            else:
                x0 = it[0].item()
                Sin.state = x0 + dx

            # Calculate perturbed solution
            blk.response()

            # Obtain all perturbed responses
            for Iout, Sout in enumerate(outps):
                # Obtain perturbed response
                fp = Sout.state

                # Finite difference sensitivity
                df = (fp - f0[Iout])/dx

                dgdx_fd = np.real(np.sum(df*df_an[Iout]))

                if dx_an[Iout][Iin] is not None:
                    try:
                        dgdx_an = np.real(dx_an[Iout][Iin][it.multi_index])
                    except (IndexError, TypeError):
                        dgdx_an = np.real(dx_an[Iout][Iin])
                else:
                    dgdx_an = 0.0

                if abs(dgdx_an) < tol:
                    error = abs(dgdx_fd - dgdx_an)
                else:
                    error = abs(dgdx_fd - dgdx_an)/max(abs(dgdx_fd), abs(dgdx_an))

                i_tested += 1
                if error > tol:
                    i_failed += 1

                if verbose or error > tol:
                    print("δ%s/δ%s     i = %s \tAn :% .3e \tFD : % .3e \tError: % .3e %s"
                          % (Sout.tag, Sin.tag, it.multi_index, dgdx_an, dgdx_fd, error, "<--*" if error > tol else ""))

                if test_fn is not None:
                    test_fn(x0, dx, dgdx_an, dgdx_fd)

            # Restore original state
            if is_iterable:
                it[0] = x0
                Sin.state = x
            else:
                Sin.state = x0

            # If the input state is complex, also do a perturbation in the imaginary direction
            if np.iscomplexobj(x0):
                # Do the perturbation
                if is_iterable:
                    it[0] += dx*1j
                else:
                    Sin.state = x0 + dx*1j

                # Calculate perturbed solution
                blk.response()

                # Obtain all perturbed responses
                for Iout, Sout in enumerate(outps):
                    # Obtain perturbed response
                    fp = Sout.state

                    # Finite difference sensitivity
                    df = (fp - f0[Iout])/(dx*1j)
                    dgdx_fd = np.imag(np.sum(df*df_an[Iout]))

                    if dx_an[Iout][Iin] is not None:
                        try:
                            dgdx_an = np.imag(dx_an[Iout][Iin][it.multi_index])
                        except IndexError:
                            dgdx_an = np.imag(dx_an[Iout][Iin])
                    else:
                        dgdx_an = 0.0

                    if abs(dgdx_an) < tol:
                        error = abs(dgdx_fd - dgdx_an)
                    else:
                        error = abs(dgdx_fd - dgdx_an)/max(abs(dgdx_fd), abs(dgdx_an))

                    i_tested += 1
                    if error > tol:
                        i_failed += 1

                    if verbose or error > tol:
                        print("δ%s/δ%s (I) i = %s \tAn :% .3e \tFD : % .3e \tError: % .3e %s"
                              % (Sout.tag, Sin.tag, it.multi_index, dgdx_an, dgdx_fd, error, "<--*" if error > tol else ""))

                    if test_fn is not None:
                        test_fn(x0, dx, dgdx_an, dgdx_fd)

                # Restore original state
                if is_iterable:
                    it[0] = x0
                else:
                    Sin.state = x0

            # Go to the next entry in the array
            it.iternext()

        print(f"-- Number of finite difference values beyond tolerance ({tol}) = {i_failed} / {i_tested}")

    print("___________________________________________________")
    print(f"\nFinished finite-difference check of \"{type(blk).__name__}\"\n")
    print("=========================================================================================================\n")


def obtain_sensitivities(signals: Iterable[Signal]) -> List:
    """ Obtains sensitivities from a list of signals, replacing None by zeros of correct length

    :param signals: The list of signals
    :return: List of sensitivities
    """
    sens = []

    for s in signals:
        the_sens = s.sensitivity

        if the_sens is None:
            the_state = s.state
            if the_state is not None:
                the_sens = np.zeros_like(the_state)

        sens.append(the_sens)
    return sens


def minimize_oc(function, variables, objective: Signal,
                tolx=1e-4, tolf=1e-4, maxit=100, xmin=0.0, xmax=1.0, move=0.2,
                l1init=0, l2init=100000, l1l2tol=1e-4, maxvol=None):
    """ Execute minimization using the OC-method

    Args:
        function: The Network defining the optimization problem
        variables: The Signals defining the design variables
        objective: The objective function Signal to be minimized

    Keyword Args:
        tolx: Stopping criterium for relative design change
        tolf: Stopping criterium for relative objective change
        maxit: Maximum number of iteration
        xmin: Minimum design variable (can be a vector)
        xmax: Maximum design variable (can be a vector)
        move: Move limit
        l1init: OC internal parameter
        l2init: OC internal parameter
        l1l2tol: OC internal parameter
    """
    xval, cumlens = _concatenate_to_array([s.state for s in variables])

    if maxvol is None:
        maxvol = np.sum(xval)

    f = 0.0
    for it in range(maxit):
        # Calculate response
        function.response()
        fprev, f = f, objective.state
        rel_fchange = abs(f-fprev)/abs(f)
        if rel_fchange < tolf:
            print(f"OC converged: Relative function change |Δf|/|f| ({rel_fchange}) below tolerance ({tolf})")
            break

        print("It. {0: 4d}, f0 = {1: .2e}, Δf = {2: .2e}".format(it, f, f-fprev))

        # Calculate sensitivity of the objective
        function.reset()
        objective.sensitivity = 1.0
        function.sensitivity()
        dfdx, _ = _concatenate_to_array(obtain_sensitivities(variables))

        # Do OC update
        l1, l2 = l1init, l2init
        while l2 - l1 > l1l2tol:
            lmid = 0.5 * (l1 + l2)
            xnew = np.clip(xval * np.sqrt(-dfdx / lmid), np.maximum(xmin, xval-move), np.minimum(xmax, xval+move))
            l1, l2 = (lmid, l2) if np.sum(xnew) - maxvol > 0 else (l1, lmid)

        # Stopping criteria on step size
        rel_stepsize = np.linalg.norm(xval - xnew)/np.linalg.norm(xval)
        if rel_stepsize < tolx:
            print(f"OC converged: Relative stepsize |Δx|/|x| ({rel_stepsize}) below tolerance ({tolx})")
            break

        xval = xnew
        # Set the new states
        for i, s in enumerate(variables):
            s.state = xnew[cumlens[i]:cumlens[i + 1]]


def minimize_mma(function, variables, responses, tolx=1e-4, tolf=1e-4, maxit=1000, xmin=0.0, xmax=1.0):
    """ Execute minimization using the MMA-method

    Uses the ``nlopt`` version of MMA.

    Todo:
        Improve the MMA implementation as the ``nlopt`` version is not very good

    Args:
        function: The Network defining the optimization problem
        variables: The Signals defining the design variables
        responses: A list of Signals, where the first is to be minimized and the others are constraints.

    Keyword Args:
        tolx: Stopping criterium for relative design change
        tolf: Stopping criterium for relative objective change
        maxit: Maximum number of iteration
        xmin: Minimum design variable (can be a vector)
        xmax: Maximum design variable (can be a vector)
    """
    # Save initial state
    xval, cumlens = _concatenate_to_array([s.state for s in variables])
    n = len(xval)

    def fi(_, grad, i):

        if grad.size > 0:
            # Calculate sensitivities
            function.reset()

            responses[i].sensitivity = 1.0

            function.sensitivity()

            grad[:], _ = _concatenate_to_array(obtain_sensitivities(variables))

        return responses[i].state

    global it
    it = 0

    # Objective function
    def f0(x, grad):
        global it
        # Set the new states
        for i, s in enumerate(variables):
            s.state = x[cumlens[i]:cumlens[i + 1]]

        # Calculate response
        function.response()

        print("It. {0: 4d}, f0 = {1: .2e}, f1 = {2: .2e}".format(it, responses[0].state, responses[1].state))
        it += 1

        return fi(x, grad, 0)

    # Create optimization
    import nlopt

    opt = nlopt.opt(nlopt.LD_MMA, n)

    opt.set_min_objective(f0)
    for ri in range(1, len(responses)):
        opt.add_inequality_constraint(lambda x, grad: fi(x, grad, ri))

    opt.set_lower_bounds(np.ones_like(xval) * xmin)
    opt.set_upper_bounds(np.ones_like(xval) * xmax)

    opt.set_maxeval(maxit)
    opt.set_xtol_rel(tolx)
    opt.set_ftol_rel(tolf)

    opt.optimize(xval)



""" Generic modules, valid for general mathematical operations """
import numpy as np
from pymoto.core_objects import Module
try:
    from opt_einsum import contract as einsum  # Faster einsum
except ModuleNotFoundError:
    from numpy import einsum


class MathGeneral(Module):
    """ General mathematical expression module

    This block can evaluate symbolic mathematical expressions. The derivatives are automatically calculated using
    ``sympy``. Variables in this expression can be given by using the signal's global name (:attr:`Signal.tag`), or by
    the signal's position in the input signal list: ``"inp0"`` for the first input signal given, ``"inp1"`` for the
    second input signal, ``"inp2"`` for the third, *etc*...

    Example:
        Two scalars::

            from pymoto import Signal, MathGeneral
            s1 = Signal("x", 1.3)
            s2 = Signal("y", 4.8)
            m = MathGeneral([s1, s2], Signal("output"), "inp0*inp1")  # or "x * y" as expression
            m.response()
            assert (m.sig_out[0].state == 1.3*4.8)

        Scalar and Vector::

            s1 = Signal("x", np.array([1.3, 2.5, 9.4]))
            s2 = Signal("y", 4.8)
            m = MathGeneral([s1, s2], Signal("output"), "sin(x)*y").response()
            assert (m.sig_out[0].state.shape == (3,))
            from math import sin
            assert (abs(m.sig_out[0].state[0] - sin(1.3)*4.8) < 1e-10)
            assert (abs(m.sig_out[0].state[1] - sin(2.5)*4.8) < 1e-10)
            assert (abs(m.sig_out[0].state[2] - sin(9.4)*4.8) < 1e-10)

    Input signals:
        ``*args`` (`float` or `np.ndarray`): Any number of numerical inputs which match the provided expression

    Output Signal:
        ``y`` (`float` or `np.ndarray`): Result of the mathematical operation

    Args:
        expression (str): The mathematical expression to be evaluated

    References:
      - `Sympy documentation <https://docs.sympy.org/latest/index.html>`_
    """
    def _prepare(self, expression):
        from sympy import lambdify
        from sympy.parsing.sympy_parser import parse_expr

        expression = expression.replace("^", "**").lower()  # Case insensitive

        # Replace powers
        expr = parse_expr(expression)

        # Variables
        var_names = []
        for i in range(len(self.sig_in)):
            var_names += ["inp{}".format(i), ]

        # Named variables <RHO, X, ...> are converted to <sig0, sig1, ...>
        trn = {}
        for i, s in enumerate(self.sig_in):
            if len(s.tag):
                if s.tag.lower() in var_names and s.tag.lower() in expression:
                    raise RuntimeError("Name '{}' multiple defined".format(s.tag.lower()))
                trn[s.tag.lower()] = var_names[i]

        expr = [expr.subs(trn)]
        self.f = lambdify(var_names, expr, "numpy")

        # Determine derivatives
        dx = []
        for i in range(len(self.sig_in)):
            dx += [expr[0].diff(var_names[i])]

        self.df = lambdify(var_names, dx, "numpy")

    def _response(self, *args):
        self.x = args
        return self.f(*args)

    def _sensitivity(self, df_dy):
        dg_df = self.df(*self.x)  # This could be moved to _response(): less computations but more memory usage

        # Initialize sensitivities with zeroed out memory. This should ensure identical type of state and sensitivity
        dg_dx = []
        for s in self.sig_in:
            try:
                dg_dxi = s.state.copy()
                dg_dxi[...] = 0
            except (AttributeError, TypeError):  # Not numpy or zero-dimension array
                dg_dxi = s.state * 0
            # assert (isinstance(dg_dxi, type(s.state)))
            dg_dx.append(dg_dxi)

        # Sum if input is scalar
        for i, sig in enumerate(self.sig_in):
            dg_dx_add = df_dy*dg_df[i]
            if np.isrealobj(dg_dx[i]) and np.iscomplexobj(dg_dx_add):
                dg_dx_add = np.real(dg_dx_add)

            # Add the contribution
            if (not hasattr(dg_dx[i], '__len__')) or (hasattr(dg_dx[i], 'ndim') and dg_dx[i].ndim == 0):
                # Scalar type or 0-dimensional array
                dg_dx[i] += np.sum(dg_dx_add)
            else:
                dg_dx[i] += dg_dx_add

            # assert (isinstance(dg_dx[i], type(sig.state)))
        return dg_dx


class EinSum(Module):
    """ General linear algebra module which uses the Numpy function ``einsum``

    Many linear algebra multiplications can be implemented using this module:

    ====================== =========================== =========================
    **Operation**          **EinSum arguments**        **Python equivalent**
    ---------------------- --------------------------- -------------------------
    Vector sum             ``"i->", u``                ``y = sum(u)``
    Elementwise multiply   ``"i,i->i", u, v``          ``w = u * v``
    Dot product            ``"i,i->", u, v``           ``y = np.dot(u,v)``
    Outer product          ``"i,j->ij", u, v``         ``A = np.outer(u,v)``
    Matrix trace           ``"ii->", A``               ``y = np.trace(A)``
    Matrix-vector product  ``"ij,j->i", A, b``         ``x = A.dot(b)``
    Quadratic form         ``"i,ij,j->", b, A, b``     ``y = b.dot(A.dot(b))``
    Matrix-matrix product  ``"ij,ij->ij", A, B``       ``C = A * B``
    Transpose matrix prod. ``"ji,ij->ij", A, B``       ``C = A.T * B``
    Matrix projection      ``"ji,jk,kl->il", V, A, V`` ``B = V.T.dot(A.dot(V))``
    ====================== =========================== =========================

    Many more advanced operations are supported (see References), with exception of expressions with repeated indices
    (*e.g.* ``iij->ij``).

    An optimized version of ``einsum`` is available by installing the package
    `opt_einsum <https://optimized-einsum.readthedocs.io/en/stable/>`_.

    Input signals:
        ``*args`` (`np.ndarray`): Any number of inputs that are passed to ``einsum`` and match the expression

    Output Signal:
        ``y`` (`np.ndarray`): Result of the ``einsum`` operation

    Args:
        expression (str): The ``einsum`` expression

    References:
      - `Numpy documentation on EinSum <https://numpy.org/doc/stable/reference/generated/numpy.einsum.html>`_
      - `Understanding Numpy's EinSum <https://stackoverflow.com/questions/26089893/understanding-numpys-einsum>`_
      - `EinStein summation in Numpy <https://obilaniu6266h16.wordpress.com/2016/02/04/einstein-summation-in-numpy/>`_
      - `Optimized Einsum opt_einsum <https://optimized-einsum.readthedocs.io/en/stable/>`_
    """
    def _prepare(self, expression: str):
        self.expr = expression
        cmd = self.expr.split("->")
        self.indices_in = [s.strip() for s in cmd[0].split(",")]
        self.indices_out = cmd[1] if "->" in self.expr else ''

    def _response(self, *args):
        return [einsum(self.expr, *args, optimize=True)]

    def _sensitivity(self, df_in):
        n_in = len(self.sig_in)

        if (self.indices_out == '') and n_in == 1:
            # In case expression has only one input and scalar output, e.g. "i->", "ij->", the output size cannot
            # be deducted. Therefore we add these exceptions
            if len(set(self.indices_in[0])) < len(self.indices_in[0]):
                # exception for repeated indices (e.g. trace, diagonal summing)
                if self.sig_in[0].state.ndim > 2:
                    raise TypeError(
                        "Sensitivities for repeated incides '{}' not supported for any other than trace 'ii->'."
                        .format(self.expr))
                mat = np.zeros_like(self.sig_in[0].state)
                np.fill_diagonal(mat, 1.0)
            else:
                mat = np.ones_like(self.sig_in[0].state)
            return np.conj(np.conj(df_in) * mat)

        for ind_in in self.indices_in:
            if len(set(ind_in)) < len(ind_in):
                raise TypeError("Sensitivities for repeated incides '{}' not supported for any other than trace 'ii->'."
                                .format(self.expr))

        df_out = []
        for ar in range(n_in):
            ind_in = [self.indices_out]
            ind_in += [elem for i, elem in enumerate(self.indices_in) if i != ar]
            arg_in = [s.state for i, s in enumerate(self.sig_in) if i != ar]
            ind_out = self.indices_in[ar]

            da_i = np.zeros_like(self.sig_in[ar].state)
            op = ",".join(ind_in)+"->"+ind_out
            einsum(op, df_in, *arg_in, out=da_i)
            df_out.append(da_i)
        return df_out



import warnings
import numpy as np
import scipy.sparse as sps
from scipy.sparse import SparseEfficiencyWarning
from .solvers import matrix_is_hermitian, matrix_is_complex, matrix_is_symmetric, LinearSolver

# ------------------------------------ Pardiso Solver -----------------------------------
try:
    from pypardiso import PyPardisoSolver
    _has_pardiso = True
except ImportError:
    _has_pardiso = False


class SolverSparsePardiso(LinearSolver):
    """ Solver wrapper Intel MKL Pardiso solver, which is a very fast and flexible multi-threaded solver

    Complex-valued matrices are currently not supported

    Uses the Python interface ``pypardiso`` to the Intel MKL PARDISO library for solving large sparse linear systems of
    equations Ax=b.

    Args:
        A (optional): The matrix
        symmetric (optional): If it is already known if the matrix is symmetric, you can provide it here
        hermitian (optional): If it is already known if the matrix is Hermitian, you can provide it here
        positive_definite (optional): If positive-definiteness is known, provide it here

    References:
        `PyPardiso <https://github.com/haasad/PyPardisoProject>`_
        `Intel MKL Pardiso <https://www.intel.com/content/www/us/en/develop/documentation/onemkl-developer-reference-c/
          top/sparse-solver-routines/onemkl-pardiso-parallel-direct-sparse-solver-iface.html>`_
    """

    defined = _has_pardiso

    def __init__(self, A=None, symmetric=None, hermitian=None, positive_definite=None):
        super().__init__(A)
        if not self.defined:
            raise ImportError("Intel MKL Pardiso solver (pypardiso) cannot be found. ")

        self.mtype = None
        self.hermitian = hermitian
        self.symmetric = symmetric
        self.positive_definite = positive_definite
        self._mtype = None
        self._pardiso_solver = None

    def _determine_mtype(self, A):
        """ Determines the matrix-type according to Intel

        ``mtype``: Matrix type
          - ``1`` Real and structurally symmetric
          - ``2`` Real and symmetric positive definite
          - ``-2`` Real and symmetric indefinite
          - ``3`` Complex and structurally symmetric
          - ``4`` Complex and Hermitian positive definite
          - ``-4`` Complex and Hermitian indefinite
          - ``6`` Complex and symmetric
          - ``11`` Real and non-symmetric
          - ``13`` Complex and non-symmetric

        References:
            `Intel MKL Pardiso arguments <https://www.intel.com/content/www/us/en/develop/documentation/
              onemkl-developer-reference-c/top/sparse-solver-routines/
              onemkl-pardiso-parallel-direct-sparse-solver-iface/pardiso.html#pardiso>`_

        Args:
            A: The matrix

        Return:
            Intel MKL mtype
        """
        # TODO structural symmetry is not checked, may be added at some point
        self.complex = matrix_is_complex(A)
        self.dtype = np.complex128 if self.complex else np.float64
        if self.complex:
            # Check if is Hermitian?
            if self.hermitian is None:
                self.hermitian = matrix_is_hermitian(A)
            if self.hermitian:
                return 4 if self.positive_definite else -4  # TODO Works if 13 is used

            # Is the matrix symmetric?
            if self.symmetric is None:
                self.symmetric = matrix_is_symmetric(A)
            return 6 if self.symmetric else 13
        else:
            # Determine symmetry
            if self.symmetric is None:
                self.symmetric = self.hermitian if self.hermitian is not None else matrix_is_symmetric(A)
            if self.symmetric:
                return 2 if self.positive_definite else -2
            return 11

    def update(self, A):
        """ Factorize the matrix A, the factorization will automatically be used if the same matrix A is passed to the
        solve method. This will drastically increase the speed of solve, if solve is called more than once for the
        same matrix A

        Args:
            A: sparse square CSR matrix (scipy.sparse.csr.csr_matrix)
        """
        # Update matrix type
        if self._mtype is None:
            self._mtype = self._determine_mtype(A)

        if self._mtype in {-2, 2, 6}:
            A = sps.triu(A, format='coo')  # Only use upper part
            # Explicitly set zero diagonal entries, as this is better for Intel Pardiso
            zero_diag_entries, = np.where(A.diagonal() == 0)
            if len(zero_diag_entries) > 0:
                A.row = np.append(A.row, zero_diag_entries)
                A.col = np.append(A.col, zero_diag_entries)
                A.data = np.append(A.data, np.zeros_like(zero_diag_entries))
            A = A.tocsr()
            # assert(np.diff(A.indptr).all())  # Every row must be non-empty
            # assert(max(np.diff(A.indices[A.indptr[:-1]])))  # All diagonals must be non-empty
        if not sps.isspmatrix_csr(A):
            warnings.warn(f'PyPardiso requires CSR matrix format, not {type(A).__name__}', SparseEfficiencyWarning)
            A = A.tocsr()
        self.A = A

        if self._pardiso_solver is None:
            self._pardiso_solver = PyPardisoSolver(mtype=self._mtype)

        self._pardiso_solver.factorize(A)

    def solve(self, b):
        """ solve Ax=b for x

        Args:
            A (scipy.sparse.csr.csr_matrix): sparse square CSR matrix , CSC matrix also possible
            b (numpy.ndarray): right-hand side(s), b.shape[0] needs to be the same as A.shape[0]

        Returns:
            Solution of the system of linear equations, same shape as input b
        """
        return self._pardiso_solver.solve(self.A, b)

    def adjoint(self, b):
        # Cannot use _pardiso_solver.solve because it changes flag 12 internally
        iparm_prev = self._pardiso_solver.get_iparm(12)
        self._pardiso_solver.set_iparm(12, int(not iparm_prev))  # Adjoint solver (transpose)
        b = self._pardiso_solver._check_b(self.A, b)
        x = self._pardiso_solver._call_pardiso(self.A, b)
        self._pardiso_solver.set_iparm(12, iparm_prev)  # Revert back to normal solver
        return x

    def _print_iparm(self):
        """ Print all iparm settings to console """
        keys = ['Undefined' for _ in self._pardiso_solver.iparm]
        keys[0] = 'Use default values'
        keys[1] = 'Fill-in reducing ordering for the input matrix'
        keys[3] = 'Preconditioned CGS/CG'
        keys[4] = 'User permutation'
        keys[5] = 'Write solution on x'
        keys[6] = 'Number of iterative refinement steps performed'
        keys[7] = 'Iterative refinement step'
        keys[8] = 'Tolerance level for the relative residual in the iterative refinement process'
        keys[9] = 'Pivoting perturbation'
        keys[10] = 'Scaling vectors'
        keys[11] = 'Solve with transposed or conjugate transposed matrix A'
        keys[12] = 'Improved accuracy using (non-) symmetric weighted matching'
        keys[13] = 'Number of perturbed pivots'
        keys[14] = 'Peak memory on symbolic factorization'
        keys[15] = 'Permanent memory on symbolic factorization'
        keys[16] = 'Size of factors/Peak memory on numerical factorization and solution'
        keys[17] = 'Report the number of non-zero elements in the factors'
        keys[18] = 'Report number of floating point operations (in 10^6 floating point operations) that are necessary ' \
                   'to factor the matrix A'
        keys[19] = 'Report CG/CGS diagnostics'
        keys[20] = 'Pivoting for symmetric indefinite matrices'
        keys[21] = 'Inertia: number of positive eigenvalues'
        keys[22] = 'Inertia: number of negative eigenvalues'
        keys[23] = 'Parallel factorization control'
        keys[24] = 'Parallel forward/backward solve control'
        keys[26] = 'Matrix checker'
        keys[27] = 'Single or double precision Intel® oneAPI Math Kernel Library PARDISO'
        keys[29] = 'Number of zero or negative pivots'
        keys[30] = 'Partial solve and computing selected components of the solution vectors'
        keys[33] = 'Optimal number of OpenMP threads for conditional numerical reproducibility (CNR) mode'
        keys[34] = 'One- or zero-based indexing of columns and rows'
        keys[35] = 'Schur complement matrix computation control'
        keys[36] = 'Format for matrix storage'
        keys[37] = 'Enable low rank update to accelerate factorization for multiple matrices with identical structure ' \
                   'and similar values'
        keys[42] = 'Control parameter for the computation of the diagonal of inverse matrix'
        keys[55] = 'Diagonal and pivoting control'
        keys[59] = 'Intel® oneAPI Math Kernel Library PARDISO mode'
        keys[62] = 'Size of the minimum OOC memory for numerical factorization and solution'
        for i, (v, k) in enumerate(zip(self._pardiso_solver.iparm, keys)):
            if v != 0:
                print(f"{i+1}: {v} ({k})")  # i+1 because of 1-based numbering


# ------------------------------------ LU Solver -----------------------------------
try:
    from scikits.umfpack import splu  # UMFPACK solver; this one is faster and has identical interface
except ImportError:
    from scipy.sparse.linalg import splu


class SolverSparseLU(LinearSolver):
    """ Solver for sparse (square) matrices using an LU decomposition.

    Internally, `scipy` uses the SuperLU library, which is relatively slow. It may be sped up using the Python package
    ``scikit-umfpack``, which scipy` is able to use, but must be installed by the user.

    References:
      - `Scipy LU <https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lu.html>`_
      - `Scipy UMFPACK <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.use_solver.html>`_
    """
    def update(self, A):
        r"""  Factorize the matrix as :math:`\mathbf{A}=\mathbf{L}\mathbf{U}`, where :math:`\mathbf{L}` is a lower
        triangular matrix and :math:`\mathbf{U}` is upper triangular.
        """
        self.iscomplex = matrix_is_complex(A)
        self.inv = splu(A)
        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{U}^{-1}\mathbf{L}^{-1}\mathbf{b}`.
        """
        return self.inv.solve(rhs)

    def adjoint(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A}^\text{H}\mathbf{x} = \mathbf{b}` by forward and
        backward substitution of :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{U}^{-\text{H}}\mathbf{b}`.
        """
        return self.inv.solve(rhs, trans=('H' if self.iscomplex else 'T'))


# ------------------------------------ Cholesky Solver scikit-sparse -----------------------------------
try:  # SCIKIT CHOLMOD
    """ Installation on Ubuntu22 using pip... Maybe it is also available in the conda repository
    sudo apt update
    sudo apt install libsuitesparse-dev
    pip install scikit-sparse
    """
    from sksparse import cholmod  # Sparse cholesky solver (CHOLMOD)
    _has_sksparse_cholmod = True
except ImportError:
    _has_sksparse_cholmod = False


class SolverSparseCholeskyScikit(LinearSolver):
    """ Solver for positive-definite Hermitian matrices using a Cholesky factorization.

    This solver requires the Python package ``scikit-sparse``. This package depends on the library ``suitesparse``,
    which can be more difficult to install on some systems. In case ``suitesparse`` cannot be installed,
    :class:`.SolverSparseCholeskyCVXOPT` is recommended, as installation of CVXOpt is easier and is packaged with
    ``suitesparse``.

    References:
      - `Scikit Installation <https://scikit-sparse.readthedocs.io/en/latest/overview.html>`_
      - `Scikit Cholmod <https://scikit-sparse.readthedocs.io/en/latest/cholmod.html>`_
    """

    defined = _has_sksparse_cholmod

    def __init__(self, A=None):
        super().__init__(A)
        if not self.defined:
            raise ImportError("scikit-sparse is not installed on this system")

    def update(self, A):
        r""" Factorize the matrix using Cholmod. In case the matrix :math:`\mathbf{A}` is non-Hermitian, the
        system of equations is solved in a least-squares sense:
        :math:`\min \left| \mathbf{A}\mathbf{x} - \mathbf{b} \right|^2`.
        The solution of this minimization is
        :math:`\mathbf{x}=(\mathbf{A}^\text{H}\mathbf{A})^{-1}\mathbf{A}^\text{H}\mathbf{b}`.
        """
        self.A = A
        if not hasattr(self, 'inv'):
            self.inv = cholmod.analyze(A)

        # Do the Cholesky factorization
        self.inv.cholesky_inplace(A)

        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{L}^{-1}\mathbf{b}` in case of an
        Hermitian matrix.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.

        """
        return self.inv(rhs)

    def adjoint(self, rhs):
        return self.solve(rhs)


# ------------------------------------ Cholesky Solver cvxopt -----------------------------------
try:  # CVXOPT cholmod
    import cvxopt
    import cvxopt.cholmod
    _has_cvxopt_cholmod = True
except ImportError:
    _has_cvxopt_cholmod = False


class SolverSparseCholeskyCVXOPT(LinearSolver):
    """ Solver for positive-definite Hermitian matrices using a Cholesky factorization.

    This solver requires the Python package ``cvxopt``.

    References:
      - `CVXOPT Installation <http://cvxopt.org/install/index.html>`_
      - `CVXOPT Cholmod <https://cvxopt.org/userguide/spsolvers.html#positive-definite-linear-equations>`_
    """

    defined = _has_cvxopt_cholmod

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.defined:
            raise ImportError("cvxopt is not installed on this system")
        self._dtype = None
        self.inv = None

    def update(self, A):
        r""" Factorize the matrix using CVXOPT's Cholmod as :math:`\mathbf{A}=\mathbf{L}\mathbf{L}^\text{H}`. """
        if not isinstance(A, cvxopt.spmatrix):
            if not isinstance(A, sps.coo_matrix):
                Kcoo = A.tocoo()
                warnings.warn(f"{type(self).__name__}: Efficiency warning: CVXOPT spmatrix must be used")
            else:
                Kcoo = A
            K = cvxopt.spmatrix(Kcoo.data, Kcoo.row.astype(int), Kcoo.col.astype(int))
        else:
            K = A

        if self.inv is None:
            self.inv = cvxopt.cholmod.symbolic(K)
        cvxopt.cholmod.numeric(K, self.inv)
        if self._dtype is None:
            self._dtype = A.dtype

        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{L}^{-1}\mathbf{b}`. """
        if rhs.dtype != self._dtype:
            warnings.warn(f"{type(self).__name__}: Type warning: rhs value type ({rhs.dtype}) is converted to {self._dtype}")
        B = cvxopt.matrix(rhs.astype(self._dtype))
        nrhs = 1 if rhs.ndim == 1 else rhs.shape[1]
        cvxopt.cholmod.solve(self.inv, B, nrhs=nrhs)
        return np.array(B).flatten() if nrhs == 1 else np.array(B)

    def adjoint(self, rhs):
        return self.solve(rhs)



import unittest
import numpy as np
import pymoto as pym
from pymoto.modules.linalg import auto_determine_solver
import sys
import inspect
np.random.seed(0)


def make_symmetric(A):
    return (A + A.T)/2


def make_hermitian(A):
    return (A + A.conj().T)/2


def make_symm_pos_def(A):
    W, V = np.linalg.eig(make_symmetric(A))
    W[W < 0.1] += abs(min(W))+0.1
    return V@np.diag(W)@(V.T)


def make_herm_pos_def(A):
    W, V = np.linalg.eigh(make_hermitian(A))
    # Eigenvalues of hermitian matrix are always real
    W[W < 0.1] += abs(min(W)) + 0.1
    return V@np.diag(W)@(V.conj().T)


""" Perpare the set of matrices we want to test """
N = 10

# Real-valued matrices
mat_real_diagonal = np.diag(np.random.rand(N))
mat_real_symm = make_symmetric(np.random.rand(N, N))
mat_real_symm_pos_def = make_symm_pos_def(np.random.rand(N, N))
mat_real_asymm = np.random.rand(N, N)

# Complex-valued matrices
mat_complex_diagonal = np.diag(np.random.rand(N) + 1j*np.random.rand(N))
mat_complex_herm = make_hermitian(np.random.rand(N, N) + 1j*np.random.rand(N, N))
mat_complex_herm_pos_def = make_herm_pos_def(np.random.rand(N, N) + 1j*np.random.rand(N, N))
mat_complex_symm = make_symmetric(np.random.rand(N, N) + 1j*np.random.rand(N, N))
mat_complex_symm_pos_def = make_symm_pos_def(np.random.rand(N, N) + 1j*np.random.rand(N, N))
mat_complex_asymm = np.random.rand(N, N) + 1j*np.random.rand(N, N)


""" ------------------ TEST UTILITY FUNCTIONS -------------------- """


class TestGenericUtility(unittest.TestCase):
    """ Generic test class for testing functions on matrix properties that return True/False """
    data = []  # The data to test: list of tuples with (matrix, property)
    fn = None  # The test function to use, which return the tested property

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if no function is given """
        if cls.fn is None:
            raise unittest.SkipTest(f"Skipping test {cls}")

    def get_tags(self):
        """ Generate tags for all the matrices to use in messaging"""
        src = inspect.getsource(type(self))
        start = src.find('data')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')

        def parse_str(t: str):
            rmv = ['\n', ' ', '(', ',', 'True', 'False']
            for c in rmv:
                t = t.replace(c, '')
            return t

        return [parse_str(t) for t in src[start:finish].strip(' []').split(')')]

    def test_all_matrices(self):
        """ Test all the matrices """
        tags = self.get_tags()
        for i, dat in enumerate(self.data):
            A = dat[0]
            res = dat[1]
            # Numpy dense matrix
            with self.subTest(msg=tags[i]):
                self.assertEqual(self.fn(A), res)


class TestIsComplex(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_complex)
    data = [
        (mat_real_diagonal, False),
        (mat_real_symm, False),
        (mat_real_symm_pos_def, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_herm, True),
        (mat_complex_herm_pos_def, True),
        (mat_complex_symm, True),
        (mat_complex_symm_pos_def, True),
        (mat_complex_asymm, True),
    ]


class TestIsDiagonal(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_diagonal)
    data = [
        (mat_real_diagonal, True),
        (mat_real_symm, False),
        (mat_real_symm_pos_def, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_herm, False),
        (mat_complex_herm_pos_def, False),
        (mat_complex_symm, False),
        (mat_complex_symm_pos_def, False),
        (mat_complex_asymm, False),
    ]


class TestIsSymmetric(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_symmetric)
    data = [
        (mat_real_diagonal, True),
        (mat_real_symm, True),
        (mat_real_symm_pos_def, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_herm, False),
        (mat_complex_herm_pos_def, False),
        (mat_complex_symm, True),
        (mat_complex_symm_pos_def, True),
        (mat_complex_asymm, False),
    ]


class TestIsHermitian(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_hermitian)
    data = [
        (mat_real_diagonal, True),
        (mat_real_symm, True),
        (mat_real_symm_pos_def, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, False),
        (mat_complex_herm, True),
        (mat_complex_herm_pos_def, True),
        (mat_complex_symm, False),
        (mat_complex_symm_pos_def, False),
        (mat_complex_asymm, False),
    ]


class TestAutoSolver(TestGenericUtility):
    """ Check if the auto_determine returns correct solver types """
    fn = staticmethod(lambda A: type(auto_determine_solver(A)))
    data = [
        (mat_real_diagonal, pym.SolverDiagonal),
        (mat_real_symm, pym.SolverDenseCholesky),
        (mat_real_symm_pos_def, pym.SolverDenseCholesky),
        (mat_real_asymm, pym.SolverDenseLU),
        (mat_complex_diagonal, pym.SolverDiagonal),
        (mat_complex_herm, pym.SolverDenseCholesky),
        (mat_complex_herm_pos_def, pym.SolverDenseCholesky),
        (mat_complex_symm, pym.SolverDenseLDL),
        (mat_complex_symm_pos_def, pym.SolverDenseLDL),
        (mat_complex_asymm, pym.SolverDenseLU),
    ]


""" ------------------ TEST THE DENSE SOLVERS -------------------- """
# Solvers:
# - SolverDiagonal
# - SolverDenseQR
# - SolverDenseLU
# - SolverDenseLDL
# - SolverDenseCholesky


class GenericTestDenseSolvers(unittest.TestCase):
    """ Generic test runner for any dense solver and list of matrices """
    solver = None  # The solver to use
    matrices = []  # The list of matrices the solver should be able to handle

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if the solver is not available """
        if cls.solver is None:
            raise unittest.SkipTest(f"No solver given for {cls.__name__}")
        if not cls.solver.defined:
            raise unittest.SkipTest(f"{type(cls.solver)} not available, lacking third-party libraries")

    def get_tags(self):
        """ Get the matrix tags for messages """
        src = inspect.getsource(type(self))
        start = src.find('matrices')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')
        return [t.strip(' ').replace('\n', '').replace(' ', '') for t in src[start:finish].strip(' []').split(',')]

    def run_solver(self, solver, A, b):
        # Get reference solutions
        Aadj = A.conj().T
        xref = np.linalg.solve(A, b)
        xrefadj = np.linalg.solve(Aadj, b)

        # Do the solution and adjoint using provided solver
        solver.update(A)
        x = solver.solve(b)
        xadj = solver.adjoint(b)

        # Check solution
        self.assertTrue(np.allclose(x, xref))
        self.assertTrue(np.allclose(A @ x - b, 0.0))

        # Check adjoint solution
        self.assertTrue(np.allclose(xadj, xrefadj))
        self.assertTrue(np.allclose(Aadj @ xadj - b, 0.0))

        # Run with LDAWrapper
        is_herm = pym.matrix_is_hermitian(A)
        is_symm = pym.matrix_is_symmetric(A)
        LDAsolver = pym.LDAWrapper(solver, hermitian=is_herm, symmetric=is_symm)
        LDAsolver.update(A)

        # First solve (do the solution)
        x = LDAsolver.solve(b)
        self.assertTrue(np.allclose(x, xref))
        self.assertTrue(np.allclose(A @ x - b, 0.0))
        self.assertTrue(LDAsolver._did_solve)  # the solution must have been done here

        # Second solve
        x = LDAsolver.solve(b)
        self.assertTrue(np.allclose(x, xref))
        self.assertTrue(np.allclose(A @ x - b, 0.0))
        self.assertFalse(LDAsolver._did_solve)  # the solution is already done

        # Adjoint solve
        xadj = LDAsolver.adjoint(b)
        self.assertTrue(np.allclose(xadj, xrefadj))
        self.assertTrue(np.allclose(Aadj @ xadj - b, 0.0))

        if is_herm or (is_symm and np.isrealobj(b)):
            self.assertFalse(LDAsolver._did_solve)
        else:
            self.assertTrue(LDAsolver._did_solve)

        # Again adjoint
        xadj = LDAsolver.adjoint(b)
        self.assertTrue(np.allclose(xadj, xrefadj))
        self.assertTrue(np.allclose(Aadj @ xadj - b, 0.0))
        self.assertFalse(LDAsolver._did_solve)  # the solution is already done

    def test_all_matrices(self):
        """ Run the tests on all given matrices """
        tags = self.get_tags()
        for A, t in zip(self.matrices, tags):
            """ Test for different type of right-hand-sides """
            sys.stdout.write(f"Test \"{self.solver.__name__}\" for matrix \"{t}\"\n")
            N = A.shape[0]
            with self.subTest(msg=f"{t}.real-rhs"):
                b = np.random.rand(N)
                self.run_solver(self.solver(), A, b)
            with self.subTest(msg=f"{t}.complex-rhs"):
                b = np.random.rand(N) + 1j * np.random.rand(N)
                self.run_solver(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-real-rhs"):
                b = np.random.rand(N, 3)
                self.run_solver(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-complex-rhs"):
                b = np.random.rand(N, 3) + 1j * np.random.rand(N, 3)
                self.run_solver(self.solver(), A, b)


class TestDenseDiagonal(GenericTestDenseSolvers):
    solver = pym.SolverDiagonal
    matrices = [
        mat_real_diagonal,
        mat_complex_diagonal,
    ]


class TestDenseQR(GenericTestDenseSolvers):
    solver = pym.SolverDenseQR
    matrices = [
        mat_real_diagonal,
        mat_real_symm,
        mat_real_symm_pos_def,
        mat_real_asymm,
        mat_complex_diagonal,
        mat_complex_herm,
        mat_complex_herm_pos_def,
        mat_complex_symm,
        mat_complex_symm_pos_def,
        mat_complex_asymm,
    ]


class TestDenseLU(GenericTestDenseSolvers):
    solver = pym.SolverDenseLU
    matrices = [
        mat_real_diagonal,
        mat_real_symm,
        mat_real_symm_pos_def,
        mat_real_asymm,
        mat_complex_diagonal,
        mat_complex_herm,
        mat_complex_herm_pos_def,
        mat_complex_symm,
        mat_complex_symm_pos_def,
        mat_complex_asymm,
    ]


class TestDenseCholesky(GenericTestDenseSolvers):
    # The indefinite matrices use LDL instead as a backup
    solver = pym.SolverDenseCholesky
    matrices = [
        mat_real_diagonal,
        mat_real_symm,
        mat_real_symm_pos_def,
        mat_complex_herm,
        mat_complex_herm_pos_def,
    ]


class TestDenseLDL(GenericTestDenseSolvers):
    solver = pym.SolverDenseLDL
    matrices = [
        mat_real_diagonal,
        mat_real_symm,
        mat_real_symm_pos_def,
        mat_complex_diagonal,
        mat_complex_herm,
        mat_complex_herm_pos_def,
        mat_complex_symm,
        mat_complex_symm_pos_def,
    ]


""" ----------- TEST LINSOLVE MODULE ---------- """


class Symm(pym.Module):
    def _response(self, A):
        return (A + A.T)/2

    def _sensitivity(self, dB):
        return (dB + dB.T)/2


class Herm(pym.Module):
    def _response(self, A):
        return (A + A.conj().T) / 2

    def _sensitivity(self, dB):
        return (dB + dB.conj().T) / 2


class TestLinSolveModule_dense(unittest.TestCase):
    # The matrices to test for
    matrices = [
        mat_real_diagonal,
        mat_real_symm,
        mat_real_symm_pos_def,
        mat_real_asymm,
        mat_complex_diagonal,
        mat_complex_herm,
        mat_complex_herm_pos_def,
        mat_complex_symm,
        mat_complex_symm_pos_def,
        mat_complex_asymm,
    ]

    def get_tags(self):
        """ Get the matrix tags for messages """
        src = inspect.getsource(type(self))
        start = src.find('matrices')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')

        def parse_str(t: str):
            rmv = ['\n', ' ', '(', ',', 'True', 'False']
            for c in rmv:
                t = t.replace(c, '')
            return t

        return [parse_str(t) for t in src[start:finish].strip(' []').split(')')]

    def run_solver(self, A, b):
        sA = pym.Signal("A", A)
        sb = pym.Signal("b", b)

        sx = pym.Signal("x")
        fn = pym.Network()
        symmetry = pym.matrix_is_symmetric(A)
        hermitian = pym.matrix_is_hermitian(A)

        if symmetry or hermitian:
            sAsys = pym.Signal("Asym", A)
            if symmetry:
                fn.append(Symm(sA, sAsys))
            elif hermitian:
                fn.append(Herm(sA, sAsys))
        else:
            sAsys = sA
        fn.append(pym.LinSolve([sAsys, sb], sx))

        fn.response()

        # Check residual
        self.assertTrue(np.allclose(sAsys.state@sx.state, sb.state))

        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sA, sb], sx, test_fn=tfn, verbose=False)

    def test_all_matrices(self):
        """ Run the tests on all given matrices """
        tags = self.get_tags()
        for A, t in zip(self.matrices, tags):
            """ Test for different type of right-hand-sides """
            sys.stdout.write(f"Test LinSolve module for matrix \"{t}\"\n")
            N = A.shape[0]
            with self.subTest(msg=f"{t}.real-rhs"):
                b = np.random.rand(N)
                self.run_solver(A, b)
            with self.subTest(msg=f"{t}.complex-rhs"):
                b = np.random.rand(N) + 1j * np.random.rand(N)
                self.run_solver(A, b)
            with self.subTest(msg=f"{t}.multi-real-rhs"):
                b = np.random.rand(N, 3)
                self.run_solver(A, b)
            with self.subTest(msg=f"{t}.multi-complex-rhs"):
                b = np.random.rand(N, 3) + 1j * np.random.rand(N, 3)
                self.run_solver(A, b)


if __name__ == '__main__':
    unittest.main()



from pymoto import Module
from .assembly import DomainDefinition
import numpy as np
from scipy.sparse import coo_matrix
from scipy.ndimage import convolve


class FilterConv(Module):
    def _prepare(self, domain: DomainDefinition, weights: np.ndarray, mode: str = 'reflect'):
        self.domain = domain
        self.weights = weights
        self.mode = mode

    def set_filter_radius(self, radius: float, relative_units: bool = False):
        if relative_units:
            dx, dy, dz = 1.0, 1.0, 1.0
        else:
            dx, dy, dz = self.domain.element_size
        delemx, delemy, delemz = int((radius-1e-10*dx)/dx), int((radius-1e-10*dy)/dy), int((radius-1e-10*dz)/dz)
        if self.domain.dim < 3:
            delemz = 0
        xrange = np.arange(-delemx, delemx+1)*dx
        yrange = np.arange(-delemy, delemy+1)*dy
        zrange = np.arange(-delemz, delemz+1)*dz
        coords_x, coords_y, coords_z = np.meshgrid(xrange, yrange, zrange)
        self.weights = np.maximum(0.0, radius - np.sqrt(coords_x*coords_x + coords_y*coords_y + coords_z*coords_z))
        self.weights /= np.sum(self.weights)  # Volume preserving
        if self.domain.dim < 3:
            self.weights = self.weights[:, :, 0]

    def _response(self, x):
        domain_sizes = [self.domain.nelx, self.domain.nely]
        if self.domain.dim >= 3:
            domain_sizes.append(self.domain.nely)
        xbox = x.reshape(*domain_sizes, order='F').T  # TODO 3d?
        return convolve(xbox, self.weights, mode=self.mode).flatten()

    def _sensitivity(self, dfdv):
        domain_sizes = [self.domain.nelx, self.domain.nely]
        if self.domain.dim == 3:
            domain_sizes.append(self.domain.nely)
        ybox = dfdv.reshape(*domain_sizes, order='F').T  # TODO 3d
        return convolve(ybox, self.weights, mode=self.mode).flatten()


class Filter(Module):
    r""" Abstract base class for any linear filter with normalization

    This module carries out the mathematical operation
    :math:`\mathbf{y} = \mathbf{S}^{-1} \mathbf{H}\mathbf{x}` in which :math:`\mathbf{S}=\text{diag}(\mathbf{s})` is a
    diagonal matrix. In index notation the same relation is written as
    :math:`y_i = \frac{\sum_j H_{ij} x_j}{ s_i }`.

    The normalization vector is the row-wise sums of :math:`\mathbf{H}`, of which the entries are calculated as
    :math:`s_i = \sum_j H_{ij}`.

    Input Signal:
        - ``x``: The unfiltered field :math:`\mathbf{x}`

    Output Signal:
        - ``y``: Filtered field :math:`\mathbf{y}`

    Keyword Args:
        nonpadding (numpy.array[int]): An array with indices at places where
          :math:`s_i = \max(\mathbf{s}) \: \forall\: i \notin \mathcal{N}`. For a density filter this mimics having values
          of `0` outside of the domain, thus emulating padding of the boundaries.
    """
    def _prepare(self, *args, nonpadding=None, **kwargs):
        self.H = self._calculate_h(*args, **kwargs).tocsc()

        self.Hs = self.H.sum(1)

        if nonpadding is not None:
            inds = ~np.isin(np.arange(len(self.Hs)), nonpadding)
            self.Hs[inds] = np.max(self.Hs)

    @staticmethod
    def _calculate_h(*args, **kwargs):
        r""" This method should be overridden by any child-classes to implement their own filtering behavior

        Returns:
            Filtering matrix :math:`\mathbf{H}`, *e.g.* in COO-format
        """
        raise NotImplementedError("Filter not implemented.")

    def _response(self, x):
        return np.asarray(self.H * x[np.newaxis].T / self.Hs)[:, 0]

    def _sensitivity(self, dfdy):
        return np.asarray(self.H * (dfdy[np.newaxis].T / self.Hs))[:, 0]


class DensityFilter(Filter):
    r""" Standard density filter for a structured mesh in topology optimization

    The filtered densities are calculated as

    :math:`y_i = \sum_j \frac{H_{ij}}{s_i}x_j`,

    where :math:`H_{ij}=\max \left( r - \sqrt{ (x_j - x_i)^2 + (y_j - y_i)^2 + (z_j - z_i)^2 } , 0 \right)`,

    and :math:`s_i=\sum_j H_{ij}`.

    Input Signal:
        - ``x``: The unfiltered field :math:`\mathbf{x}`

    Output Signal:
        - ``y``: Filtered field :math:`\mathbf{y}`

    Args:
        domain: The domain layout

    Keyword Args:
        radius (float or int): The filtering radius (in absolute units of elements)
        nonpadding (numpy.array[int]): An array with indices at places where
          :math:`s_i = \max(\mathbf{s}) \: \forall\: i \notin \mathcal{N}`. For a density filter this mimics having
          values of ``0`` outside of the domain, thus emulating padding of the boundaries.

    References:
      - Bruns & Tortorelli (2001). *Topology optimization of non-linear elastic structures and compliant mechanisms*.
        Computer Methods in Applied Mechanics and Engineering, 190(26–27), 3443–3459.
        `doi: 10.1016/S0045-7825(00)00278-4 <https://doi.org/10.1016/S0045-7825(00)00278-4>`_
      - Bourdain (2001). *Filters in topology optimization*. International Journal for Numerical Methods in
        Engineering, 50, 2143-2158. `doi: 10.1002/nme.116 <https://doi.org/10.1002/nme.116>`_
    """

    @staticmethod
    def _calculate_h(domain: DomainDefinition, radius=2.0):
        """ Density filter: Build (and assemble) the index+data vectors for the coo matrix format
        Total number of filter entries - for every element, a number of entries wrt other elements are needed (not
        including boundaries)
        Takes a square domain around current element, encompassing circle with radius floor(rmin)
        """
        delem = int(radius)
        # Number of elements
        nx, ny, nz = domain.nelx, domain.nely, max(domain.nelz, 1)
        nel = domain.nel  # Number of elements

        xrange = np.arange(0, nx)
        yrange = np.arange(0, ny)
        zrange = np.arange(0, nz)

        # Get element grid with indices in x, y, and z directions
        xinds, yinds, zinds = np.meshgrid(xrange, yrange, zrange, indexing='ij')

        # Obtain global element numbers
        els = domain.get_elemnumber(xinds, yinds, zinds)

        # Rearrange directional indices according to element number
        ix = np.zeros(nel, dtype=int)
        iy = np.zeros(nel, dtype=int)
        iz = np.zeros(nel, dtype=int)
        ix[els] = xinds
        iy[els] = yinds
        iz[els] = zinds

        # Determine the limits of the window wrt current element
        xlow = np.maximum(ix - delem, 0)
        xupp = np.minimum(ix + delem, nx - 1)
        ylow = np.maximum(iy - delem, 0)
        yupp = np.minimum(iy + delem, ny - 1)
        zlow = np.maximum(iz - delem, 0)
        zupp = np.minimum(iz + delem, nz - 1)

        # Number of window elements in x, y and z direction and the total number per element
        nwindx = xupp - xlow + 1
        nwindy = yupp - ylow + 1
        nwindz = zupp - zlow + 1
        nwind = nwindx * nwindy * nwindz

        # Total number of window elements
        ncum = np.cumsum(nwind)

        nfilter = ncum[-1]

        # Initialize sparse indices for coo format
        h_rows = np.zeros(nfilter, dtype=np.uint32)
        h_cols = np.zeros(nfilter, dtype=np.uint32)

        # For all elements, get the surrounding elements
        for el in range(nel):
            elcomp = els[xlow[el]:xupp[el]+1, ylow[el]:yupp[el]+1, zlow[el]:zupp[el]+1].flatten()

            indstart = ncum[el-1] if el > 0 else 0
            h_rows[indstart:ncum[el]] = el
            h_cols[indstart:ncum[el]] = elcomp

        # Calculate element distances
        dx = ix[h_rows] - ix[h_cols]
        dy = iy[h_rows] - iy[h_cols]
        dz = iz[h_rows] - iz[h_cols]

        h_values = np.maximum(0.0, radius - np.sqrt(dx*dx + dy*dy + dz*dz))

        # Finalize assembly
        return coo_matrix((h_values, (h_rows, h_cols)), shape=(nel, nel))


class OverhangFilter(Module):
    r""" Implementation of overhang filter by Langelaar (2016, 2017)

    It proceeds layer by layer through the entire domain. For each element in the current layer, the maximum printable
    density is determined by a smooth maximum of the supporting elements
    :math:`s_i = \text{smax}(\mathbf{y}_\text{supp})`. Then, the final printed density value is obtained by a smooth
    minimum operation of the desired density
    :math:`x_i` and the maximum printable density :math:`x_i` as :math:`y_i = \text{smin}(x_i, s_i)`.

    Input Signal:
        - ``x``: The unfiltered field :math:`\mathbf{x}`

    Output Signal:
        - ``y``: Filtered field :math:`\mathbf{y}`, without overhangs

    Args:
        domain: The domain layout

    Keyword Args:
        direction: Print direction as array or string, e.g. ``[0, -1]`` (in 2D) or ``"y-"`` for negative y direction.
          Currently, only directions aligned with one of the Cartesian axes are supported. Default is ``[0, 1, 0]``
        xi_0: Density value for which zero overshoot is required ( ``0 <= xi_0 <= 1`` ). Default is ``0.5``
        p: Exponent of the smooth maximum function ( ``p > 0`` ). Higher p increases accuracy, but reduces smoothness.
          Default is ``40.0``
        eps: Smooth minimum regularization parameter ( ``eps >= 0`` ). Lower eps increases accuracy, but reduces
          smoothness. Default is ``1e-4``
        nsampling: ``3`` for 2D overhang, ``5`` or ``9`` for 3D overhang. Default is ``3`` in 2D and ``5`` in 3D

    References:
      - Langelaar, M. (2017). *An additive manufacturing filter for topology optimization of print-ready designs*.
        Structural and Multidisciplinary Optimization, 55(3), 871–883.
        `doi: 10.1007/s00158-016-1522-2 <https://doi.org/10.1007/s00158-016-1522-2>`_
      - Langelaar, M. (2016). *Topology optimization of 3D self-supporting structures for additive manufacturing*.
        Additive Manufacturing, 12, 60–70.
        `doi: 10.1016/j.addma.2016.06.010 <https://doi.org/10.1016/j.addma.2016.06.010>`_
    """
    def _prepare(self,
                 domain: DomainDefinition,
                 direction=(0.0, 1.0, 0.0),
                 xi_0: float = 0.5,
                 p: float = 40.0,
                 eps: float = 1e-4,
                 nsampling: int = None):

        # Parse print direction
        if isinstance(direction, str):
            # Print axis
            axes = np.argwhere([a in direction.lower() for a in ['x', 'y', 'z']]).flatten()
            if axes.size != 1:
                raise ValueError(f"Wronly specified print direction {direction}, should be e.g. \"+x\", \"-y\"")

            # Print direction
            direction = [0.0, 0.0, 0.0]
            direction[axes[0]] = -1.0 if '-' in direction else +1.0
        direction = np.asarray(direction, dtype=np.float64).flatten()
        if direction.size < 3:
            direction = np.pad(direction, (0, 3-direction.size), 'constant', constant_values=0.0)
        elif direction.size > 3:
            direction = direction[:3]

        self.direction = direction / np.linalg.norm(direction)
        self.domain = domain
        if self.domain.dim == 2:
            assert self.direction[2] == 0.0, "Z-direction must be zero for 2-dimensional domain"
        assert abs(self.direction).sum() >= 1.0 - 1e-10, "The print direction must be aligned with either x, y or z directions"

        if nsampling is None:
            nsampling = 3 if self.domain.dim == 2 else 5
        assert (self.domain.dim == 2 and nsampling == 3) or (self.domain.dim == 3 and (nsampling == 5 or nsampling == 9)), \
            "Only 3 (2D), 5, or 9 (3D) support points supported"

        # Parameters
        self.xi_0 = xi_0
        self.p = p
        self.eps = eps
        self.nsampling = nsampling
        self.q, self.shift, self.backshift = None, None, None
        self.smax = None

    def set_parameters(self, typ: np.dtype):
        """ Set the internal smooth-maximum and smooth-minimum parameters according to the values in Langelaar, 2017

        Args:
            typ: The numeric type used in the calculations
        """
        dbl_min = np.finfo(typ).tiny
        self.q = self.p + np.log(1.0*self.nsampling) / np.log(self.xi_0)
        self.shift = 100.0 * pow(dbl_min, 1.0/self.p)  # Small shift to prevent division by 0
        self.backshift = pow(self.nsampling, 1/self.q)*pow(self.shift, self.p/self.q)*0.95  # 5% smaller to be on the safe side

    def _response(self, x):
        if self.q is None:  # Set parameters according to data type of x
            self.set_parameters(x.dtype)
        xprint = x.copy()
        self.smax = x.copy()

        # Size of the domain
        size = [self.domain.nelx, self.domain.nely, max(self.domain.nelz, 1)]

        dir_layer = int(np.argmax(abs(self.direction)))  # The axis of the print direction
        dx_layer = int(np.sign(self.direction[dir_layer]))  # Iteration direction
        ind_layer = 1 if dx_layer >= 0 else size[dir_layer]-2  # Starting index

        dir_orth1 = (dir_layer + 1) % 3
        dir_orth2 = (dir_layer + 2) % 3
        if dir_orth1 == 2 and self.domain.dim == 2:
            dir_orth1, dir_orth2 = dir_orth2, dir_orth1  # Make sure the z-direction is last for 2D

        #  Select layer offsets from:
        #                |<-- 3x 2D supports -->|<- 5-point 3D ->|<--         9-point 3D          -->|
        layer_offsets = [[-1, 0], [0, 0], [1, 0], [0, -1], [0, 1], [-1, -1], [-1, 1], [1, -1], [1, 1]][:self.nsampling]

        entire_layer = np.meshgrid(range(size[dir_orth1]), range(size[dir_orth2]), indexing='ij')
        # Create masks since not all elements may be inside the domain
        support_idx = [None for _ in range(self.nsampling)]
        offset_masks = [None for _ in range(self.nsampling)]
        for i, offset in enumerate(layer_offsets):
            support_idx[i] = (entire_layer[0] + offset[0], entire_layer[1] + offset[1])
            offset_masks[i] = np.logical_and((support_idx[i][0] >= 0) * (support_idx[i][0] < size[dir_orth1]),
                                             (support_idx[i][1] >= 0) * (support_idx[i][1] < size[dir_orth2]))

        # Loop over all the layers
        while 0 <= ind_layer < size[dir_layer]:
            # 1) Get all support values
            keep = np.zeros_like(entire_layer[0], dtype=x.dtype)
            for i, (supp_idx, supp_mask) in enumerate(zip(support_idx, offset_masks)):
                el = [None, None, None]
                el[dir_layer] = ind_layer-dx_layer
                el[dir_orth1] = supp_idx[0][supp_mask]
                el[dir_orth2] = supp_idx[1][supp_mask]
                els = self.domain.get_elemnumber(*el)
                keep[supp_mask] += np.power(xprint[els]+self.shift, self.p)

            # 2) Take smooth maximum
            max_supp = np.power(keep, 1/self.q)-self.backshift
            # max_supp = np.maximum.reduce(supp_vals)  # Absolute maximum

            # 3) Take smooth minimum
            el = [None, None, None]
            el[dir_layer] = ind_layer
            el[dir_orth1] = entire_layer[0]
            el[dir_orth2] = entire_layer[1]
            els = self.domain.get_elemnumber(*el)
            self.smax[els] = max_supp  # Save maximum printable densities for sensitivity
            r1 = x[els] - max_supp
            xprint[els] = (x[els] + max_supp - np.sqrt(r1*r1 + self.eps) + np.sqrt(self.eps))/2
            # xprint[els] = np.minimum(x[els], max_supp)

            ind_layer += dx_layer

        return xprint

    def _sensitivity(self, dxprint):
        x = self.sig_in[0].state
        xprint = self.sig_out[0].state
        dx = np.zeros_like(dxprint)

        # Size of the domain
        size = [self.domain.nelx, self.domain.nely, max(self.domain.nelz, 1)]

        dir_layer = int(np.argmax(abs(self.direction)))  # The axis of the print direction
        dx_layer = int(np.sign(self.direction[dir_layer]))  # Iteration direction
        ind_layer = size[dir_layer]-1 if dx_layer >= 0 else 0  # Starting index (="ending" in response)

        dir_orth1 = (dir_layer + 1) % 3
        dir_orth2 = (dir_layer + 2) % 3
        if dir_orth1 == 2 and self.domain.dim == 2:
            dir_orth1, dir_orth2 = dir_orth2, dir_orth1  # Make sure the z-direction is last for 2D

        #  Select layer offsets from:
        #                |<-- 3x 2D supports -->|<- 5-point 3D ->|<--         9-point 3D          -->|
        layer_offsets = [[-1, 0], [0, 0], [1, 0], [0, -1], [0, 1], [-1, -1], [-1, 1], [1, -1], [1, 1]][:self.nsampling]

        entire_layer = np.meshgrid(range(size[dir_orth1]), range(size[dir_orth2]), indexing='ij')
        support_idx = [None for _ in range(self.nsampling)]
        offset_masks = [None for _ in range(self.nsampling)]
        for i, offset in enumerate(layer_offsets):
            support_idx[i] = (entire_layer[0] + offset[0], entire_layer[1] + offset[1])
            offset_masks[i] = np.logical_and((support_idx[i][0] >= 0) * (support_idx[i][0] < size[dir_orth1]),
                                             (support_idx[i][1] >= 0) * (support_idx[i][1] < size[dir_orth2]))

        # Loop over all the layers
        while True:
            # 3) Take smooth minimum
            el = [None, None, None]
            el[dir_layer] = ind_layer
            el[dir_orth1] = entire_layer[0]
            el[dir_orth2] = entire_layer[1]
            els = self.domain.get_elemnumber(*el)

            # xprint[els] = (x[els] + max_supp - np.sqrt(r1*r1 + self.eps) + np.sqrt(self.eps))/2
            r1 = x[els] - self.smax[els]
            dfdr1 = -dxprint[els] * r1 / (2*np.sqrt(r1*r1+self.eps))
            dx[els] = dxprint[els]/2 + dfdr1
            dfdsmax = dxprint[els]/2 - dfdr1

            # 2) Take smooth maximum
            # max_supp = np.power(keep, 1/self.q)-self.backshift
            keep = np.power(self.smax[els] + self.backshift, self.q)
            dfdkeep = dfdsmax * np.power(keep, (1/self.q) - 1) / self.q
            c = self.p*dfdkeep

            # 1) Get all support values
            for i, (supp_idx, supp_mask) in enumerate(zip(support_idx, offset_masks)):
                el = [None, None, None]
                el[dir_layer] = ind_layer-dx_layer
                el[dir_orth1] = supp_idx[0][supp_mask]
                el[dir_orth2] = supp_idx[1][supp_mask]
                els = self.domain.get_elemnumber(*el)
                #  keep[supp_mask] += np.power(xprint[els]+self.shift, self.p)
                dxprint[els] += c[supp_mask]*np.power(xprint[els]+self.shift, self.p-1)

            ind_layer -= dx_layer
            if not 1 <= ind_layer < size[dir_layer]-1:
                break

        # Base layer is directly transferred
        el = [None, None, None]
        el[dir_layer] = ind_layer
        el[dir_orth1] = entire_layer[0]
        el[dir_orth2] = entire_layer[1]
        els = self.domain.get_elemnumber(*el)
        dx[els] = dxprint[els]
        return dx



import warnings
import numpy as np
import scipy.linalg as spla  # Dense matrix solvers
from .solvers import matrix_is_hermitian, matrix_is_diagonal, LinearSolver


class SolverDiagonal(LinearSolver):
    """ Solver for diagonal matrices """
    def update(self, A):
        """ Extracts the diagonal of the matrix """
        self.diag = A.diagonal()
        return self

    def solve(self, rhs):
        r""" Solve using the diagonal only, by :math:`x_i = b_i / A_{ii}`

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        if rhs.ndim == 1:
            return rhs / self.diag
        else:
            return rhs / self.diag[..., None]

    def adjoint(self, rhs):
        r""" Solve using the diagonal only, by :math:`x_i = b_i / A_{ii}^*`

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        return self.solve(rhs.conj()).conj()


# Dense QR solver
class SolverDenseQR(LinearSolver):
    """ Solver for dense (square) matrices using a QR decomposition """
    def update(self, A):
        r""" Factorize the matrix as :math:`\mathbf{A}=\mathbf{Q}\mathbf{R}`, where :math:`\mathbf{Q}` is orthogonal
        (:math:`\mathbf{Q}^\text{H}=\mathbf{Q}^{-1}`) and :math:`\mathbf{R}` is upper triangular.
        """
        self.A = A
        self.q, self.r = spla.qr(A)
        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by backward substitution of
        :math:`\mathbf{x} = \mathbf{R}^{-1}\mathbf{Q}^\text{H}\mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        return spla.solve_triangular(self.r, self.q.T.conj()@rhs)

    def adjoint(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A}^\text{H}\mathbf{x} = \mathbf{b}` by
        forward substitution of :math:`\mathbf{x} = \mathbf{Q}\mathbf{R}^{-H}\mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        return self.q@spla.solve_triangular(self.r, rhs, trans='C')


# Dense LU solver
class SolverDenseLU(LinearSolver):
    """ Solver for dense (square) matrices using an LU decomposition """
    def update(self, A):
        r"""  Factorize the matrix as :math:`\mathbf{A}=\mathbf{L}\mathbf{U}`, where :math:`\mathbf{L}` is a lower
        triangular matrix and :math:`\mathbf{U}` is upper triangular.
        """
        self.p, self.l, self.u = spla.lu(A)
        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{U}^{-1}\mathbf{L}^{-1}\mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        return spla.solve_triangular(self.u, spla.solve_triangular(self.l, self.p.T@rhs, lower=True))

    def adjoint(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A}^\text{H}\mathbf{x} = \mathbf{b}` by forward and
        backward substitution of :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{U}^{-\text{H}}\mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        return self.p@spla.solve_triangular(self.l, spla.solve_triangular(self.u, rhs, trans='C'),
                                            lower=True, trans='C')  # TODO permutation


# Dense Cholesky solver
class SolverDenseCholesky(LinearSolver):
    """ Solver for Hermitian positive-definite matrices using a Cholesky factorization.
    In case the matrix is singular and factorization fails, a backup-solver is used (:class:`.SolverDenseLDL`).
    """
    def __init__(self, *args, **kwargs):
        self.backup_solver = SolverDenseLDL()
        self.success = None
        super().__init__(*args, **kwargs)

    def update(self, A):
        r""" Factorize the matrix as :math:`\mathbf{A}=\mathbf{U}^{\text{H}}\mathbf{U}`, where :math:`\mathbf{U}` is an
        upper triangular matrix.
        """
        try:
            self.u = spla.cholesky(A)
            self.success = True
        except np.linalg.LinAlgError as err:
            warnings.warn(f"{type(self).__name__}: {err} -- using {type(self.backup_solver).__name__} instead")
            self.backup_solver.update(A)
            self.success = False
        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{U}^{-1}\mathbf{U}^{-\text{H}}\mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        # TODO When Cholesky factorization A = U^T U is used, symmetric complex matrices can also be solved, but this is
        #  not implemented in scipy
        if self.success:
            return spla.solve_triangular(self.u, spla.solve_triangular(self.u, rhs, trans='C'))
        else:
            return self.backup_solver.solve(rhs)

    def adjoint(self, rhs):
        r""" A Hermitian matrix is self-adjoint (:math:`\mathbf{A}=\mathbf{A}^\text{H}`), so this is equal to the
        regular solution.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        if self.success:
            return self.solve(rhs)
        else:
            return self.backup_solver.adjoint(rhs)


# Dense LDL solver
class SolverDenseLDL(LinearSolver):
    """ Solver for Hermitian or symmetric matrices using an LDL factorization. Unlike :class:`.SolverDenseCholesky`,
    it is able to factorize (some, not all) indefinite matrices, as well as symmetric complex (thus non-Hermitian)
    matrices.

    Requires scipy>=1.7
    """
    def __init__(self, *args, hermitian=None, **kwargs):
        self.hermitian = hermitian
        super().__init__(*args, **kwargs)

    def update(self, A):
        r""" Factorize the matrix as :math:`\mathbf{A}=\mathbf{L}\mathbf{D}\mathbf{L}^{\text{H}}` in case it is
        Hermitian, or as :math:`\mathbf{A}=\mathbf{L}\mathbf{D}\mathbf{L}^{\text{T}}` if it is symmetric. In the case
        matrix :math:`\mathbf{A}` is real-valued, there is no difference between the two.
        The matrix :math:`\mathbf{L}` is lower triangular and :math:`\mathbf{D}` is a diagonal matrix.
        """
        if self.hermitian is None:
            self.hermitian = matrix_is_hermitian(A)
        self.l, self.d, self.p = spla.ldl(A, hermitian=self.hermitian)  # LDL is introduced in Scipy v1.7
        if matrix_is_diagonal(self.d):  # Exact diagonal
            d1 = np.diag(1/np.diag(self.d))
        else:
            d1 = np.linalg.inv(self.d)  # This could be improved by looking at blocks on the diagonal
        self.dinv = lambda b: d1@b
        self.dinvH = lambda b: (d1.conj().T)@b
        self.lp = self.l[self.p, :]
        return self

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by forward and backward
        substitution of :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{D}^{-1}\mathbf{L}^{-1}\mathbf{b}` in the
        Hermitian case or as :math:`\mathbf{x} = \mathbf{L}^{-\text{T}}\mathbf{D}^{-1}\mathbf{L}^{-1}\mathbf{b}` in the
        symmetric case.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        u1 = spla.solve_triangular(self.lp, rhs[self.p], lower=True, unit_diagonal=True)
        u2 = self.dinv(u1)
        u = np.zeros_like(rhs, dtype=u2.dtype)
        u[self.p] = spla.solve_triangular(self.lp, u2, trans='C' if self.hermitian else 'T', lower=True, unit_diagonal=True)
        return u

    def adjoint(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A}^\text{H} \mathbf{x} = \mathbf{b}` by forward and
        backward substitution of
        :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{D}^{-\text{H}}\mathbf{L}^{-1}\mathbf{b}` in the  Hermitian
        case or as :math:`\mathbf{x} = \mathbf{L}^{-\text{H}}\mathbf{D}^{-\text{H}}\mathbf{L}^{-*}\mathbf{b}`
        in the symmetric case.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        if not self.hermitian:
            u1 = spla.solve_triangular(self.lp, rhs[self.p].conj(), lower=True, unit_diagonal=True).conj()
        else:
            u1 = spla.solve_triangular(self.lp, rhs[self.p], lower=True, unit_diagonal=True)
        u2 = self.dinvH(u1)
        u = np.zeros_like(rhs, dtype=u2.dtype)
        u[self.p] = spla.solve_triangular(self.lp, u2, trans='C', lower=True, unit_diagonal=True)
        return u



from typing import Union, Iterable
import warnings
import numpy as np
from numpy.typing import NDArray
from scipy.sparse import spmatrix
from ..utils import _parse_to_list
try:  # Import fast optimized einsum
    from opt_einsum import contract as einsum
except ModuleNotFoundError:
    from numpy import einsum


def isdyad(x):
    """ Checks if argument is a ``DyadCarrier`` """
    return isinstance(x, DyadCarrier)


def isdense(x):
    """ Checks if argument is a dense ``numpy`` array """
    return isinstance(x, np.ndarray)


def isscalarlike(x):
    """ Checks if argument is either a scalar, an array scalar, or a 0-dim array """
    return np.isscalar(x) or (isdense(x) and x.ndim == 0)


def isnullslice(x):
    """ Checks if argument is ``:`` slice """
    return isinstance(x, slice) and x == slice(None, None, None)


class DyadCarrier(object):
    """ Efficient storage for dyadic or rank-N matrix

    Stores only the vectors instead of creating a full rank-N matrix
    :math:`\mathbf{A} = \sum_k^N \mathbf{u}_k\otimes\mathbf{v}_k`
    or in index notation :math:`A_{ij} = \sum_k^N u_{ki} v_{kj}`. This saves a lot of memory for low :math:`N`.

    Args:
        u : (optional) List of vectors
        v : (optional) List of vectors (if ``u`` is given and ``v`` not, a symmetric dyad is assumed with ``v = u``)
    """

    __array_priority__ = 11.0  # For overriding numpy's ufuncs
    ndim = 2  # Number of dimensions

    def __init__(self, u: Iterable = None, v: Iterable = None):
        self.u = []
        self.v = []
        self.ulen = -1
        self.vlen = -1
        self.dtype = np.dtype('float64')  # Standard data type
        self.add_dyad(u, v)

    @property
    def shape(self):
        """ The shape of the matrix (nrows, ncols) """
        return (self.ulen, self.vlen)

    @property
    def size(self):
        """ Size of the matrix (nrows x ncols) """
        if self.ulen < 0 or self.vlen < 0:
            return 0
        else:
            return self.ulen * self.vlen

    def add_dyad(self, u: Iterable, v: Iterable = None, fac: float = None):
        """ Adds a list of vectors to the dyad carrier

        Checks for conforming sizes of `u` and `v`. The data inside the vectors are copied.

        It is possible to add higher-dimensional matrices, which are summed along all but the last dimension.
        The effect of this is that also cross-terms are added, e.g.
        adding the matrices :math:`\mathbf{U}=[\mathbf{u}_1, \mathbf{u}_2, \mathbf{u}_3]`
        and :math:`\mathbf{V}=[\mathbf{v}_1, \mathbf{v}_2, \mathbf{v}_3]`
        results in :math:`(\mathbf{u}_1+\mathbf{u}_2+\mathbf{u}_3)\otimes(\mathbf{v}_1+\mathbf{v}_2+\mathbf{v}_3)`
        being added to the DyadCarrier.

        Args:
            u: List of vectors
            v: List of vectors [Optional: If not given, use `v=u`]
            fac: Optional multiplication factor

        Returns:
            self
        """
        ulist = _parse_to_list(u)
        vlist = ulist if v is None else _parse_to_list(v)

        if len(ulist) != len(vlist):
            raise TypeError("Number of vectors in u ({}) and v({}) should be equal".format(len(ulist), len(vlist)))

        n = len(ulist)

        for i, ui, vi in zip(range(n), ulist, vlist):
            # Make sure they are numpy arrays
            if not isinstance(ui, np.ndarray):
                ui = np.array(ui)
            if not isinstance(vi, np.ndarray):
                vi = np.array(vi)

            # Sum if it is a block-matrix U.V^T = sum u_i.v_j^T
            if ui.ndim > 1:
                ui = ui.sum(axis=tuple(range(ui.ndim - 1)))
            elif ui.ndim == 0:
                ui = ui[np.newaxis]

            if vi.ndim > 1:
                vi = vi.sum(axis=tuple(range(vi.ndim - 1)))
            elif vi.ndim == 0:
                vi = vi[np.newaxis]

            # Update the dyadic matrix dimensions
            if self.ulen < 0:
                self.ulen = ui.shape[-1]

            if self.vlen < 0:
                self.vlen = vi.shape[-1]

            # Check dimensions
            if ui.shape[-1] != self.ulen:
                raise TypeError(f"U vector {i} of shape {ui.shape}, not conforming to dyad size {self.ulen}.")

            if vi.shape[-1] != self.vlen:
                raise TypeError(f"V vector {i} of shape {vi.shape}, not conforming to dyad size {self.vlen}.")

            # Don't add zero vectors
            if np.linalg.norm(ui) == 0 or np.linalg.norm(vi) == 0:
                continue

            # Add the vectors
            self.u.append(ui.copy() if fac is None else fac*ui)
            self.v.append(vi.copy())

            # Update the type
            self.dtype = np.result_type(self.dtype, ui.dtype)
            self.dtype = np.result_type(self.dtype, vi.dtype)
        return self

    def __getitem__(self, subscript):
        assert len(subscript) == self.ndim, "Invalid number of slices, must be 2"
        usub = [ui[subscript[0]] for ui in self.u]
        vsub = [vi[subscript[1]] for vi in self.v]
        subdyad = DyadCarrier(usub, vsub)
        if isscalarlike(usub[0]) or isscalarlike(vsub[0]):
            mat = subdyad.todense()
            if isscalarlike(usub[0]) and isscalarlike(vsub[0]):
                return mat[0, 0]
            else:
                return mat.flatten()
        elif isinstance(subscript[0], np.ndarray) and isinstance(subscript[1], np.ndarray):
            if subscript[0].shape != subscript[1].shape:
                raise IndexError(f"shape mismatch: indexing arrays could not be broadcast together "
                                 f"with shapes {subscript[0].shape} {subscript[1].shape}")
            return subdyad.diagonal()
        else:
            return subdyad

    def __setitem__(self, subscript, value):
        assert len(subscript) == self.ndim, "Invalid number of slices, must be 2"
        if value != 0.0:
            raise ValueError("Setting entries to other than 0 makes no sense")
        if not isnullslice(subscript[0]) and not isnullslice(subscript[1]):
            raise IndexError("Only full-column or full-row slices can be set, e.g. [:, 3] or [3:8, :]")
        for ui, vi in zip(self.u, self.v):
            if not isnullslice(subscript[0]):
                ui[subscript[0]] = value
            if not isnullslice(subscript[1]):
                vi[subscript[1]] = value

    def __pos__(self):
        return DyadCarrier(self.u, self.v)

    def __neg__(self):
        return DyadCarrier([-uu for uu in self.u], self.v)

    def __iadd__(self, other):
        self.add_dyad(other.u, other.v)
        return self

    def __add__(self, other):  # self + other
        if isscalarlike(other):
            if other == 0:
                return self.copy()
            raise NotImplementedError('adding a nonzero scalar from a '
                                      'dyadcarrier is not supported')
        elif isdyad(other):
            if other.shape != self.shape and (self.size > 0 and other.size > 0):
                raise ValueError(f"Inconsistent shapes {self.shape} and {other.shape}")
            return DyadCarrier(self.u, self.v).__iadd__(other)
        elif isdense(other):
            other = np.broadcast_to(other, self.shape)
            return other + self.todense()
        else:
            return NotImplemented

    def __radd__(self, other):  # other + self
        return self.__add__(other)

    def __isub__(self, other):
        self.add_dyad(other.u, other.v, fac=-1.0)
        return self

    def __sub__(self, other):  # self - other
        return self.__add__(-other)

    def __rsub__(self, other):  # other - self
        if isscalarlike(other):
            if other == 0:
                return -self.copy()
            raise NotImplementedError('subtracting a dyadcarrier from a '
                                      'nonzero scalar  is not supported')
        elif isdense(other):
            other = np.broadcast_to(other, self.shape)
            return other - self.todense()
        else:
            return NotImplemented

    def __rmul__(self, other):  # other * self
        return DyadCarrier([other*ui for ui in self.u], self.v)

    def __mul__(self, other):  # self * other
        return DyadCarrier(self.u, [vi*other for vi in self.v])

    def copy(self):
        """ Returns a deep copy of the DyadCarrier """
        return DyadCarrier(self.u, self.v)

    def conj(self):
        """ Returns (a deep copied) complex conjugate of the DyadCarrier """
        return DyadCarrier([u.conj() for u in self.u], [v.conj() for v in self.v])

    @property
    def real(self):
        """ Returns a deep copy of the real part of the DyadCarrier """
        return DyadCarrier([*[u.real for u in self.u], *[-u.imag for u in self.u]], [*[v.real for v in self.v], *[v.imag for v in self.v]])

    @property
    def imag(self):
        """ Returns a deep copy of the imaginary part of the DyadCarrier """
        return DyadCarrier([*[u.real for u in self.u], *[u.imag for u in self.u]], [*[v.imag for v in self.v], *[v.real for v in self.v]])

    # flake8: noqa: C901
    def contract(self, mat: Union[NDArray, spmatrix] = None, rows: NDArray[int] = None, cols: NDArray[int] = None):
        r""" Performs a number of contraction operations using the DyadCarrier

        Calculates the result(s) of the quadratic form:
        :math:`y = \sum_k \mathbf{u}_k^{\text{T}} \mathbf{B} \mathbf{v}_k`

        Examples:
            * Contraction using identity matrix, is the sum of dot products, which is equal to the trace of the rank-N
              matrix:

              ``y = A.contract()`` equals
              :math:`y = \text{trace}(\mathbf{A}) = \sum_k \mathbf{u}_k^{\text{T}}\mathbf{v}_k`

            * Contraction using matrix ``B`` of size ``(N, M)`` calculates the quadratic form:

              ``y = A.contract(B)`` equals
              :math:`y = \sum_k \mathbf{u}_k^{\text{T}} \mathbf{B} \mathbf{v}_k`

            * Sliced contraction with sliced row, matrix ``B`` of size ``(n, M)`` and ``rows`` of size ``n``:

              ``y = A.contract(B, rows)``
              :math:`y =\sum_k \mathbf{u}[\texttt{rows}]_k^{\text{T}} \mathbf{B} \mathbf{v}_k`

            * Sliced contraction with sliced column, matrix ``B`` of size ``(N, m)`` and ``cols`` of size ``m``:

              ``y = A.contract(B, cols=cols)`` equals
              :math:`y = \sum_k \mathbf{u}_k^{\text{T}} \mathbf{B} \mathbf{v}[\texttt{cols}]_k`

            * Sliced contraction with both sliced rows and columns, matrix ``B`` of size ``(n, m)``,
              ``rows`` of size ``n``, and ``cols`` of size ``m``:

              ``y = A.contract(B, rows, cols)`` equals
              :math:`y = \sum_k \mathbf{u}[\texttt{rows}]_k^{\text{T}} \mathbf{B} \mathbf{v}[\texttt{cols}]_k`

            * Batch contraction with multiple matrices in batch mode, matrix ``B`` of size ``(P, N, M)``:

              ``y = A.contract(B)`` equals
              :math:`y_p = \sum_k \mathbf{u}_k^{\text{T}} \mathbf{B}_p \mathbf{v}_k`

            * Batch contraction with multiple slices, matrix ``B`` of size ``(n, M)``, ``rows`` of size ``(P, n)``:

              ``y = A.contract(B, rows)`` equals
              :math:`y_p = \sum_k \mathbf{u}[\texttt{rows}_p]_k^{\text{T}} \mathbf{B} \mathbf{v}_k`

            * Batch contraction with multiple matrices and slices, matrix ``B`` of size ``(P, n, M)``, ``rows`` of
              size ``(P, n)``:

              ``y = A.contract(B, rows)`` equals
              :math:`y_p = \sum_k \mathbf{u}[\texttt{rows}_p]_k^{\text{T}} \mathbf{B}_p \mathbf{v}_k`

            * Batch contraction with multiple matrices, row slice, and column slice, which is used ,`e.g.`, for finite
              element sensitivities. Matrix ``B`` is of size ``(P, n, m)``, ``rows`` of  size ``(P, n)``, and ``cols``
              of size ``(P, m)``:

              ``y = A.contract(B, rows, cols)`` equals
              :math:`y_p = \sum_k \mathbf{u}[\texttt{rows}_p]_k^{\text{T}} \mathbf{B}_p \mathbf{v}[\texttt{cols}_p]_k`

            * Batch contractions can be extended with multiple extra batch-dimensions. For instance, matrix
              ``B`` of size ``(P, Q, R, N, M)`` results in ``y`` of size ``(P, Q, R)``:

              ``y = A.contract(B)`` equals
              :math:`y_{pqr} = \sum_k \mathbf{u}_k^{\text{T}} \mathbf{B}_{pqr} \mathbf{v}_k`

        Args:
            mat: The matrix to contract with (optional)
            rows: Indices for the rows (optional)
            cols: Indices for the columns to use (optional)

        Returns:
            Contraction result
        """

        rowvar = 'i'
        colvar = 'i' if mat is None else 'j'
        matvar = 'ij'

        # Batch variables
        isbatchmat = mat is not None and mat.ndim > 2
        isbatchrow = rows is not None and rows.ndim > 1
        isbatchcol = cols is not None and cols.ndim > 1

        batchsize = mat.shape[:-2] if isbatchmat else None

        if isbatchrow:
            if batchsize is None:
                batchsize = rows.shape[:-1]
            elif rows.shape[:-1] != batchsize:
                raise ValueError("Batch size of rows {} not conforming to {}".format(rows.shape[:-1], batchsize))

        if isbatchcol:
            if batchsize is None:
                batchsize = cols.shape[:-1]
            elif cols.shape[:-1] != batchsize:
                raise ValueError("Batch size of cols {} not conforming to {}".format(cols.shape[:-1], batchsize))

        if batchsize is None:
            val = 0.0
            for ui, vi in zip(self.u, self.v):
                uarg = ui if rows is None else ui[rows]
                varg = vi if cols is None else vi[cols]
                if mat is None:
                    val += uarg @ varg
                else:
                    val += uarg @ mat @ varg
            return val

        # Continue in batch mode
        batchvar = ''.join([chr(i+65) for i in range(len(batchsize))])

        if isbatchmat:
            matvar = batchvar + matvar

        if isbatchrow:
            rowvar = batchvar + rowvar

        if isbatchcol:
            colvar = batchvar + colvar

        exprvars = (rowvar, colvar) if mat is None else (rowvar, matvar, colvar)
        expr = ','.join(exprvars) + '->' + batchvar

        val = 0.0 if batchsize is None else np.zeros(batchsize)
        for ui, vi in zip(self.u, self.v):
            uarg = ui if rows is None else ui[rows]
            varg = vi if cols is None else vi[cols]
            argums = (uarg, varg) if mat is None else (uarg, mat, varg)
            val += einsum(expr, *argums)

        return val

    def todense(self):
        """ Returns a full (dense) matrix from the DyadCarrier matrix """
        warning_size = 100e+6  # Bytes
        if (self.shape[0]*self.shape[1]*self.dtype.itemsize) > warning_size:
            warnings.warn(f"Expanding a dyad results into a dense matrix. "
                          f"This is not advised for large matrices {self.shape}", ResourceWarning, stacklevel=2)

        val = np.zeros((max(0, self.shape[0]), max(0, self.shape[1])), dtype=self.dtype)

        for ui, vi in zip(self.u, self.v):
            val += np.outer(ui, vi)

        return val

    def iscomplex(self):
        """ Check if the DyadCarrier is of complex type """
        return np.iscomplexobj(np.array([], dtype=self.dtype))

    def diagonal(self, k: int = 0):
        """ Returns the diagonal of the DyadCarrier matrix """
        if (self.shape[0] == 0) or (self.shape[1] == 0):
            return np.zeros(0, dtype=self.dtype)

        ustart = max(0, -k)
        vstart = max(0,  k)
        n = min(self.shape[0]-ustart, self.shape[1]-vstart)

        if n < 0:
            return np.zeros(0, dtype=self.dtype)

        diag = np.zeros(n, dtype=self.dtype)

        for ui, vi in zip(self.u, self.v):
            diag += ui[ustart:ustart+n] * vi[vstart:vstart+n]

        return diag

    @property
    def T(self):
        """ Shorthand transpose (returns deep copy) """
        return self.transpose()

    def transpose(self):
        """ Returns a deep copy of the transposed DyadCarrier matrix"""
        return DyadCarrier(self.v, self.u)

    def dot(self, other):
        """ Inner product """
        return self.__dot__(other)

    def __dot__(self, other):  # self.dot(other)
        if other.ndim == 2:
            return self.__matmul__(other)

        val = np.zeros_like(self.u[0])
        for ui, vi in zip(self.u, self.v):
            val += ui * vi.dot(other)
        return val

    def __rdot__(self, other):  # other.dot(self)
        if other.ndim == 2:
            return self.__rmatmul__(other)

        val = np.zeros_like(self.v[0])
        for ui, vi in zip(self.u, self.v):
            val += vi * other.dot(ui)
        return val

    def __matmul__(self, other):  # self @ other
        if other.ndim == 1:
            return self.__dot__(other)

        return DyadCarrier(self.u, [vi@other for vi in self.v])

    def __rmatmul__(self, other):  # other @ self
        if other.ndim == 1:
            return self.__rdot__(other)
        return DyadCarrier([other@ui for ui in self.u], self.v)



import unittest
import pymoto as pym
import numpy as np
np.random.seed(0)


class ComplexVecDot(pym.Module):
    def _response(self, u, v):
        return u@v

    def _sensitivity(self, dy):
        u, v = [s.state for s in self.sig_in]
        return dy*v, dy*u


class TestComplex(unittest.TestCase):
    def test_real_to_real_via_complex(self):
        N = 4
        s_ur, s_ui = pym.Signal('ur', np.random.rand(N)), pym.Signal('ui', np.random.rand(N))
        s_vr, s_vi = pym.Signal('vr', np.random.rand(N)), pym.Signal('vi', np.random.rand(N))

        ma = pym.MakeComplex([s_ur, s_ui], pym.Signal("uc"))
        mb = pym.MakeComplex([s_vr, s_vi], pym.Signal("vc"))
        m1 = ComplexVecDot([ma.sig_out[0], mb.sig_out[0]], pym.Signal('uv'))
        m2 = pym.ComplexNorm(m1.sig_out, pym.Signal("|uv|"))

        pym.finite_difference(pym.Network(ma, mb, m1, m2), [s_ur, s_ui, s_vr, s_vi], m2.sig_out, dx=1e-5)

    def test_real_to_complex(self):
        N = 4
        s_ur, s_ui = pym.Signal('ur', np.random.rand(N)), pym.Signal('ui', np.random.rand(N) * 0)
        s_vr, s_vi = pym.Signal('vr', np.random.rand(N)), pym.Signal('vi', np.random.rand(N) * 0)

        ma = pym.MakeComplex([s_ur, s_ui], pym.Signal("uc"))
        mb = pym.MakeComplex([s_vr, s_vi], pym.Signal("vc"))
        m1 = ComplexVecDot([ma.sig_out[0], mb.sig_out[0]], pym.Signal('u.v'))

        pym.finite_difference(pym.Network(ma, mb, m1), [s_ur, s_ui, s_vr, s_vi], m1.sig_out, dx=1e-5, random=False,
                              keep_zero_structure=False)

    def test_real_to_complex1(self):
        N = 4
        s_ur, s_ui = pym.Signal('ur', np.random.rand(N)), pym.Signal('ui', np.random.rand(N))
        ma = pym.MakeComplex([s_ur, s_ui], pym.Signal("uc"))
        pym.finite_difference(ma, dx=1e-5, random=False)

    def test_complex_to_real(self):
        s_z = pym.Signal('z', np.array([0.3 + 0.4*1j]))
        m1 = pym.RealPart(s_z, pym.Signal('ur'))
        m2 = pym.ImagPart(s_z, pym.Signal('ui'))
        pym.finite_difference(pym.Network(m1, m2), s_z, [*m1.sig_out, *m2.sig_out], dx=1e-5, random=False)

    def test_complex_to_real1(self):
        s_z = pym.Signal('z', np.array([0.3 + 0.4*1j]))
        m = pym.ComplexNorm(s_z, pym.Signal('|z|'))
        pym.finite_difference(m, dx=1e-5, random=False)

    def test_complex_to_real2(self):
        N = 4
        u = np.random.rand(N) + 1j * np.random.rand(N)
        v = np.random.rand(N) + 1j * np.random.rand(N)
        m1 = ComplexVecDot([pym.Signal('u', u), pym.Signal('v', v)], pym.Signal('u'))
        m2 = pym.ComplexNorm(m1.sig_out, pym.Signal("ur"))
        pym.finite_difference(pym.Network(m1, m2), m1.sig_in, m2.sig_out, dx=1e-5)

    def test_complex_to_complex(self):
        N = 4
        u = np.random.rand(N) + 1j * np.random.rand(N)
        v = np.random.rand(N) + 1j * np.random.rand(N)
        m1 = ComplexVecDot([pym.Signal('u', u), pym.Signal('v', v)], pym.Signal('u.v'))
        pym.finite_difference(m1, dx=1e-5)



import numpy as np
import scipy.sparse as sps
try:
    import cvxopt
    _has_cvxopt = True
except ImportError:
    _has_cvxopt = False


def is_cvxopt_spmatrix(A):
    """ Checks if the argument is a cvxopt sparse matrix """
    return isinstance(A, cvxopt.spmatrix) if _has_cvxopt else False


def matrix_is_complex(A):
    """ Checks if the matrix is complex """
    if is_cvxopt_spmatrix(A):
        return A.typecode == 'z'
    else:
        return np.iscomplexobj(A)


def matrix_is_diagonal(A):
    """ Checks if the matrix is diagonal"""
    if sps.issparse(A):
        if isinstance(A, sps.dia_matrix):
            return len(A.offsets) == 1 and A.offsets[0] == 0
        else:
            return np.allclose((A - sps.spdiags(A.diagonal(), 0, *A.shape)).data, 0.0)
    elif is_cvxopt_spmatrix(A):
        return max(abs(A.I - A.J)) == 0
    else:
        return np.allclose(A, np.diag(np.diag(A)))


def matrix_is_symmetric(A):
    """ Checks whether a matrix is numerically symmetric """
    if sps.issparse(A):
        return np.allclose((A-A.T).data, 0)
    elif is_cvxopt_spmatrix(A):
        return np.isclose(max(abs(A-A.T)), 0.0)
    else:
        return np.allclose(A, A.T)


def matrix_is_hermitian(A):
    """ Checks whether a matrix is numerically Hermitian """
    if matrix_is_complex(A):
        if sps.issparse(A):
            return np.allclose((A-A.T.conj()).data, 0)
        elif is_cvxopt_spmatrix(A):
            return np.isclose(max(abs(A-A.ctrans())), 0.0)
        else:
            return np.allclose(A, A.T.conj())
    else:
        return matrix_is_symmetric(A)


class LinearSolver:
    """ Base class of all linear solvers

    Keyword Args:
        A (matrix): Optionally provide a matrix, which is used in :method:`update` right away.

    Attributes:
        defined (bool): Flag if the solver is able to run, e.g. false if some dependent library is not available
    """

    defined = True
    _err_msg = ""

    def __init__(self, A=None):
        if A is not None:
            self.update(A)

    def update(self, A):
        """ Updates with a new matrix of the same structure

        Args:
            A (matrix): The new matrix of size ``(N, N)``

        Returns:
            self
        """
        raise NotImplementedError(f"Solver not implemented {self._err_msg}")

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}`

        Args:
            rhs: Right hand side :math:`\mathbf{b}` of shape ``(N)`` or ``(N, K)`` for multiple right-hand-sides

        Returns:
            Solution vector :math:`\mathbf{x}` of same shape as :math:`\mathbf{b}`
        """
        raise NotImplementedError(f"Solver not implemented {self._err_msg}")

    def adjoint(self, rhs):
        r""" Solves the adjoint linear system of equations

        The system of equations is :math:`\mathbf{A}^\text{H} \mathbf{x} = \mathbf{b}` (conjugate transpose) in case of
        complex matrix or :math:`\mathbf{A}^\text{T} \mathbf{x} = \mathbf{b}` for a real-valued matrix.

        Args:
            rhs: Right hand side :math:`\mathbf{b}` of shape ``(N)`` or ``(N, K)`` for multiple right-hand-sides

        Returns:
            Solution vector :math:`\mathbf{x}` of same shape as :math:`\mathbf{b}`
        """
        raise NotImplementedError(f"Solver not implemented {self._err_msg}")

    @staticmethod
    def residual(A, x, b):
        r""" Calculates the (relative) residual of the linear system of equations

        The residual is calculated as
        :math:`r = \frac{\left| \mathbf{A} \mathbf{x} - \mathbf{b} \right|}{\left| \mathbf{b} \right|}`

        Args:
            A: The matrix
            x: Solution vector
            b: Right-hand side

        Returns:
            Residual value
        """
        return np.linalg.norm(A@x - b) / np.linalg.norm(b)


class LDAWrapper(LinearSolver):
    r""" Linear dependency aware solver (LDAS)

    This solver uses previous solutions of the system :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` to reduce computational
    effort. In case the solution :math:`\mathbf{x}` is linearly dependent on the previous solutions, the solution
    will be nearly free of cost.

    Args:
        solver: The internal solver to be used
        tol (optional): Residual tolerance above which the internal solver is used to add a new solution vector.
        A (optional): The matrix :math:`\mathbf{A}`

    References:
        Koppen, van der Kolk, van den Boom, Langelaar (2022).
        `doi: 10.1007/s00158-022-03378-8 <https://doi.org/10.1007/s00158-022-03378-8>`_
    """
    def __init__(self, solver: LinearSolver, tol=1e-8, A=None, hermitian=False, symmetric=False):
        self.solver = solver
        self.tol = tol
        self.x_stored = []
        self.b_stored = []
        self.xadj_stored = []
        self.badj_stored = []
        self.A = None
        self._did_solve = False  # For debugging purposes
        self._last_rtol = 0.
        self.hermitian = hermitian
        self.symmetric = symmetric
        super().__init__(A)

    def update(self, A):
        """ Clear the internal stored solution vectors and update the internal ``solver`` """
        self.A = A
        self.x_stored.clear()
        self.b_stored.clear()
        self.xadj_stored.clear()
        self.badj_stored.clear()
        self.solver.update(A)

    def _do_solve_1rhs(self, A, rhs, x_data, b_data, solve_fn):
        rhs_loc = rhs.copy()
        sol = 0

        # Check linear dependencies in the rhs using modified Gram-Schmidt
        for (x, b) in zip(x_data, b_data):
            alpha = rhs_loc.conj() @ b / (b.conj() @ b)
            rhs_loc -= alpha * b
            sol += alpha * x

        # Check tolerance
        self._last_rtol = 1.0 if len(x_data) == 0 else self.residual(A, sol, rhs)

        if self._last_rtol > self.tol:
            # Calculate a new solution
            xnew = solve_fn(rhs_loc)
            x_data.append(xnew)
            b_data.append(rhs_loc)
            sol += xnew
            self._did_solve = True
        else:
            self._did_solve = False

        return sol

    def _solve_1x(self, b):
        return self._do_solve_1rhs(self.A, b, self.x_stored, self.b_stored, self.solver.solve)

    def _adjoint_1x(self, b):
        return self._do_solve_1rhs(self.A.conj().T, b, self.xadj_stored, self.badj_stored, self.solver.adjoint)

    def solve(self, rhs):
        r""" Solves the linear system of equations :math:`\mathbf{A} \mathbf{x} = \mathbf{b}` by performing a modified
        Gram-Schmidt over the previously calculated solutions :math:`\mathbf{U}` and corresponding right-hand-sides
        :math:`\mathbf{F}`. This is used to construct an approximate solution
        :math:`\tilde{\mathbf{x}} = \sum_k \alpha_k \mathbf{u}_k` in the subspace of :math:`\mathbf{U}`.
        If the residual of :math:`\mathbf{A} \tilde{\mathbf{x}} = \mathbf{b}` is above the tolerance, a new solution
        :math:`\mathbf{u}_{k+1}` will be added to the database such that
        :math:`\mathbf{x} = \tilde{\mathbf{x}}+\mathbf{u}_{k+1}` is the solution to the system
        :math:`\mathbf{A} \mathbf{x} = \mathbf{b}`.

        The right-hand-side :math:`\mathbf{b}` can be of size ``(N)`` or ``(N, K)``, where ``N`` is the size of matrix
        :math:`\mathbf{A}` and ``K`` is the number of right-hand sides.
        """
        if rhs.ndim == 1:
            return self._solve_1x(rhs)
        else:  # Multiple rhs
            sol = []
            for i in range(rhs.shape[-1]):
                sol.append(self._solve_1x(rhs[..., i]))
            return np.stack(sol, axis=-1)

    def adjoint(self, rhs):
        if self.hermitian:
            return self.solve(rhs)
        elif self.symmetric:
            return self.solve(rhs.conj()).conj()
        else:
            if rhs.ndim == 1:
                return self._adjoint_1x(rhs)
            else:  # Multiple rhs
                sol = []
                for i in range(rhs.shape[-1]):
                    sol.append(self._adjoint_1x(rhs[..., i]))
                return np.stack(sol, axis=-1)



import os
import sys
import base64
import struct
import warnings
from typing import Union
import numpy as np


class DomainDefinition:
    r""" Definition for a structured domain
    Nodal numbering used in the domain is given below.

    Quadrangle in 2D

    ::

              ^
              | y
        3 -------- 4
        |     |    |   x
        |      --- | ---->
        |          |
        1 -------- 2


    Hexahedron in 3D

    ::

               y
        2----------3
        |\     ^   |\
        | \    |   | \
        |  \   |   |  \
        |   6------+---7
        |   |  +-- |-- | -> x
        0---+---\--1   |
         \  |    \  \  |
          \ |     \  \ |
           \|      z  \|
            4----------5


    Args:
        nelx : Number of elements in x-direction
        nely : Number of elements in y-direction
        nelz : Number of elements in z-direction; if zero it is a 2D domain
        unitx : Element size in x-direction
        unity : Element size in y-direction
        unitz : Element size in z-direction

    Attributes:
        dim : Dimensionality of the object
        nel : Total number of elements
        nnodes : Total number of nodes
        elemnodes : Number of nodes per element
        node_numbering : The numbering scheme used to number the nodes in each element
        conn : Connectivity matrix of size (# elements, # nodes per element)
    """

    def __init__(self, nelx: int, nely: int, nelz: int = 0, unitx: float = 1.0, unity: float = 1.0, unitz: float = 1.0):
        self.nelx, self.nely, self.nelz = nelx, nely, nelz
        if self.nely is None:
            self.nely = 0
        if self.nelz is None:
            self.nelz = 0
        self.unitx, self.unity, self.unitz = unitx, unity, unitz

        self.dim = 1 if (self.nelz == 0 and self.nely == 0) else (2 if self.nelz == 0 else 3)

        self.element_size = np.array([unitx, unity, unitz])
        assert np.prod(self.element_size[:self.dim]) > 0.0, 'Element volume needs to be positive'

        self.nel = self.nelx * self.nely * max(self.nelz, 1)  # Total number of elements
        self.nnodes = (self.nelx + 1) * (self.nely + 1) * (self.nelz + 1)  # Total number of nodes

        self.elemnodes = 2 ** self.dim  # Number of nodes in each element

        # This is where the node numbering is defined, users may override this in their program to use custom numbering
        self.node_numbering = [[0, 0, 0] for _ in range(self.elemnodes)]
        self.node_numbering[0] = [-1, -1, -1]
        self.node_numbering[1] = [+1, -1, -1]
        if self.dim >= 2:
            self.node_numbering[2] = [-1, +1, -1]
            self.node_numbering[3] = [+1, +1, -1]
        if self.dim >= 3:
            self.node_numbering[4] = [-1, -1, +1]
            self.node_numbering[5] = [+1, -1, +1]
            self.node_numbering[6] = [-1, +1, +1]
            self.node_numbering[7] = [+1, +1, +1]

        # Get the element numbers
        elx = np.repeat(np.arange(self.nelx), self.nely * max(self.nelz, 1))
        ely = np.tile(np.repeat(np.arange(self.nely), max(self.nelz, 1)), self.nelx)
        elz = np.tile(np.arange(max(self.nelz, 1)), self.nelx * self.nely)
        el = self.get_elemnumber(elx, ely, elz)

        # Setup node-element connectivity
        self.conn = np.zeros((self.nel, self.elemnodes), dtype=int)
        self.conn[el, :] = self.get_elemconnectivity(elx, ely, elz)

    def get_elemnumber(self, eli: Union[int, np.ndarray], elj: Union[int, np.ndarray], elk: Union[int, np.ndarray] = 0):
        """ Gets the element number(s) for element(s) with given Cartesian indices (i, j, k)

        Args:
            eli : Ith element in the x-direction; can be integer or array
            elj : Jth element in the y-direction; can be integer or array
            elk : Kth element in the z-direction; can be integer or array

        Returns:
            The element number(s) corresponding to selected indices
        """
        return (elk * self.nely + elj) * self.nelx + eli

    def get_nodenumber(self, nodi: Union[int, np.ndarray], nodj: Union[int, np.ndarray], nodk: Union[int, np.ndarray] = 0):
        """ Gets the node number(s) for nodes with given Cartesian indices (i, j, k)

        Args:
            nodi : Ith node in the x-direction; can be integer or array
            nodj : Jth node in the y-direction; can be integer or array
            nodk : Kth node in the z-direction; can be integer or array

        Returns:
            The node number(s) corresponding to selected indices
        """
        return (nodk * (self.nely + 1) + nodj) * (self.nelx + 1) + nodi

    def get_elemconnectivity(self, i: Union[int, np.ndarray], j: Union[int, np.ndarray], k: Union[int, np.ndarray] = 0):
        """ Get the connectivity for element identified with Cartesian indices (i, j, k)
        This is where the nodal numbers are defined

        Args:
            i: Ith element in the x-direction; can be integer or array
            j: Jth element in the y-direction; can be integer or array
            k: Kth element in the z-direction; can be integer or array

        Returns:
            The node numbers corresponding to selected elements of size (# selected elements, # nodes per element)
        """
        nods = [self.get_nodenumber(i+max(n[0], 0), j+max(n[1], 0), k+max(n[2], 0)) for n in self.node_numbering]
        return np.stack(nods, axis=-1)

    def get_dofconnectivity(self, ndof: int):
        """ Get the connectivity in terms of degrees of freedom

        Args:
            ndof: The number of degrees of freedom per node

        Returns:
            The dof numbers corresponding to each element of size (# total elements, # dofs per element)
        """
        return np.repeat(self.conn*ndof, ndof, axis=-1) + np.tile(np.arange(ndof), self.elemnodes)

    def eval_shape_fun(self, pos: np.ndarray):
        r""" Evaluate the linear shape functions of the finite element

        In 1D
          * :math:`N_1(x) = \frac{1}{w} \left(\frac{w}{2} - x\right)`
          * :math:`N_2(x) = \frac{1}{w} \left(\frac{w}{2} + x\right)`

        In 2D [1]
          * :math:`N_1(x,y) = \frac{1}{A} \left(\frac{w}{2} - x\right) \left(\frac{h}{2} - y\right)`
          * :math:`N_2(x,y) = \frac{1}{A} \left(\frac{w}{2} + x\right) \left(\frac{h}{2} - y\right)`
          * :math:`N_3(x,y) = \frac{1}{A} \left(\frac{w}{2} - x\right) \left(\frac{h}{2} + y\right)`
          * :math:`N_4(x,y) = \frac{1}{A} \left(\frac{w}{2} + x\right) \left(\frac{h}{2} + y\right)`

        with :math:`A = wh`

        In 3D
          * :math:`N_1(x,y,z) = \frac{1}{V} \left(\frac{w}{2} - x\right) \left(\frac{h}{2} - y\right) \left(\frac{d}{2} - z\right)`
          * etc.

        with :math:`V = whd`

        Args:
            pos : Evaluation coordinates [x, y, z (optional)] within bounds of [-element_size/2, element_size/2]

        Returns:
            Array of evaluated shape functions [N1(x), N2(x), ...]

        References:
            [1] Cook, Malkus, Plesha, Witt (2002). Concepts and applications of finite element analysis (4th ed.), eq. (6.2-3)
        """
        v = np.prod(self.element_size[:self.dim])
        assert v > 0.0, 'Element volume needs to be positive'
        ret = np.ones(self.nnodes)/v
        for i in range(self.dim):
            ret *= np.array([self.element_size[i] + n[i]*pos[i] for n in self.node_numbering])
        return ret

    def eval_shape_fun_der(self, pos: np.ndarray):
        """ Evaluates the shape function derivatives in x, y, and optionally z-direction.
        For 1D domains, the y and z directions are optional.
        For 2D domains, the z direction is optional.

        Args:
            pos : Evaluation coordinates [x, y, z(optional)] within bounds of [-element_size/2, element_size/2]

        Returns:
            Shape function derivatives of size (#dimensions, #shape functions)
        """
        v = np.prod(self.element_size[:self.dim])
        assert v > 0.0, 'Element volume needs to be positive'
        dN_dx = np.ones((self.dim, self.elemnodes))/v  # dN/dx = 1/V
        for i in range(self.dim):
            for j in range(self.dim):
                if i != j:  # dN/dx_i *= (w[j]/2 ± x[j])
                    dN_dx[i, :] *= np.array([self.element_size[j]/2 + n[j]*pos[j] for n in self.node_numbering])
            dN_dx[i, :] *= np.array([n[i] for n in self.node_numbering])  # Flip +/- signs according to node position
        return dN_dx

    # flake8: noqa: C901
    def write_to_vti(self, vectors: dict, filename="out.vti", scale=1.0, origin=(0.0, 0.0, 0.0)):
        """ Write all given vectors to a Paraview (VTI) file

        The size of the vectors should be a multiple of ``nel`` or ``nnodes``. Based on their size they are marked as
        cell-data or point-data in the VTI file. For 2D data (size is equal to ``2*nnodes``), the z-dimension is padded
        with zeros to have 3-dimensional data. Also block-vectors of multiple dimensions (*e.g.* ``(2, 3*nnodes)``) are
        accepted, which get the suffixed as ``_00``.

        Args:
            vectors: A dictionary of vectors to write. Keys are used as vector names.
            filename (str): The file loction
            scale: Uniform scaling of the gridpoints
            origin: Origin of the domain
        """
        ext = '.vti'
        if ext not in os.path.splitext(filename)[-1].lower():
            filename += ext

        # Sort into point-data and cell-data
        point_dat = {}
        cell_dat = {}
        for key, vec in vectors.items():
            if vec.size % self.nel == 0:
                cell_dat[key] = vec
            elif vec.size % self.nnodes == 0:
                point_dat[key] = vec
            else:
                warnings.warn(f"Vector {key} is neither cell- nor point-data. Skipping vector...")

        if len(point_dat) == 0 and len(cell_dat) == 0:
            warnings.warn(f"Nothing to write to {filename}. Skipping file...")
            return

        len_enc = ('<' if sys.byteorder == 'little' else '>') + 'Q'

        with open(filename, 'wb') as file:
            # XML header
            file.write(b'<?xml version=\"1.0\"?>\n')

            # Vtk header
            byte_order = "LittleEndian" if sys.byteorder == "little" else "BigEndian"
            file.write(f'<VTKFile type=\"ImageData\" version=\"0.1\" '
                       f'header_type=\"UInt64\" '
                       f'byte_order=\"{byte_order}\">\n'.encode())

            # Extend of coordinates
            file.write(f"<ImageData WholeExtent=\"0 {self.nelx} 0 {self.nely} 0 {self.nelz}\"".encode())

            # Origin of domain
            file.write(f" Origin=\"{origin[0]*scale} {origin[1]*scale} {origin[2]*scale}\"".encode())

            # Spacing of points (dx, dy, dz)
            dx, dy, dz = self.element_size[0:3]*scale
            file.write(f" Spacing=\"{dx} {dy} {dz}\">\n".encode())

            # Start new piece
            file.write(f"<Piece Extent=\"0 {self.nelx} 0 {self.nely} 0 {self.nelz}\">\n".encode())

            # Start writing pointdata
            if len(point_dat) > 0:
                file.write(b'<PointData>\n')
                for key, vec in point_dat.items():
                    vecax = next((i for i, s in enumerate(vec.shape) if s % self.nnodes == 0), None)
                    ncomponents = vec.shape[vecax]//self.nnodes
                    # Vectorize 2D vectors, by padding with 0's
                    pad_to_vector = ncomponents == 2 and self.dim == 2
                    assert vec.ndim <= 2, "Only for 1D and 2D numpy arrays"
                    nvectors = 1 if vec.ndim == 1 else vec.shape[(vecax+1) % 2]
                    nzeros = int(np.ceil(np.log10(nvectors)))
                    for i in range(nvectors):
                        vecname = key
                        if nvectors > 1:
                            intstr = i.__format__(f'0{nzeros}d')
                            vecname += f"({intstr})"
                            ind = [slice(None), slice(None)]
                            ind[(vecax+1) % 2] = i
                            vec_to_write = vec[tuple(ind)].astype(np.float32)
                        else:
                            vec_to_write = vec.astype(np.float32)

                        if pad_to_vector:
                            vec_pad = np.zeros(3*self.nnodes, dtype=np.float32)
                            vec_pad[0::3] = vec_to_write[0::2]
                            vec_pad[1::3] = vec_to_write[1::2]
                            vec_to_write = vec_pad

                        file.write(f'<DataArray type=\"Float32\" '
                                   f'Name=\"{vecname}\" '
                                   f'NumberOfComponents=\"{3 if pad_to_vector else ncomponents}\" '
                                   f'format=\"binary\">\n'.encode())
                        enc_data = base64.b64encode(vec_to_write)  # Encode the data
                        # Get the length of encoded data block
                        enc_len = base64.b64encode(struct.pack(len_enc, len(enc_data)))
                        file.write(enc_len)  # Write length
                        file.write(enc_data)  # Write data
                        file.write(b'\n</DataArray>\n')
                file.write(b'</PointData>\n')

            # Start writing celldata
            if len(cell_dat) > 0:
                file.write(b'<CellData>\n')
                for key, vec in cell_dat.items():
                    vecax = next((i for i, s in enumerate(vec.shape) if s % self.nel == 0), None)
                    ncomponents = vec.shape[vecax] // self.nel
                    assert vec.ndim <= 2, "Only for 1D and 2D numpy arrays"
                    nvectors = 1 if vec.ndim == 1 else vec.shape[(vecax+1) % 2]

                    for i in range(nvectors):
                        vecname = key
                        if nvectors > 1:
                            vecname += f"({i})"
                            ind = [slice(None), slice(None)]
                            ind[(vecax+1) % 2] = i
                            vec_to_write = vec[tuple(ind)].astype(np.float32)
                        else:
                            vec_to_write = vec.astype(np.float32)

                        file.write(f'<DataArray type=\"Float32\" '
                                   f'Name=\"{vecname}\" '
                                   f'NumberOfComponents=\"{ncomponents}\" '
                                   f'format=\"binary\">\n'.encode())
                        enc_data = base64.b64encode(vec_to_write)  # Encode the data
                        # Get the length of encoded data block
                        enc_len = base64.b64encode(struct.pack(len_enc, len(enc_data)))
                        file.write(enc_len)  # Write length
                        file.write(enc_data)  # Write data
                        file.write(b'\n</DataArray>\n')
                file.write(b'</CellData>\n')

            file.write(b'</Piece>\n')
            file.write(b'</ImageData>\n')
            file.write(b'</VTKFile>')



import unittest
import numpy as np
import pymoto as pym
np.random.seed(0)


class Symm(pym.Module):
    def _response(self, A):
        return (A + A.T)/2

    def _sensitivity(self, dB):
        return (dB + dB.T)/2


class Herm(pym.Module):
    def _response(self, A):
        return (A + A.T.conj())/2

    def _sensitivity(self, dB):
        return (dB + dB.T.conj())/2


class TestEigenSolverDense(unittest.TestCase):
    def use_solver(self, A, symmetry=False, hermitian=False):
        sA = pym.Signal("A", A)

        slambda = pym.Signal("lambda")
        sU = pym.Signal("eigvec")

        fn = pym.Network()
        if symmetry or hermitian:
            if hermitian and symmetry and np.iscomplexobj(A):
                raise RuntimeError("Complex matrix cannot be symmetric and hermitian at the same time")
            Modifier = Herm if hermitian else Symm
            sAsys = pym.Signal("Asym", A)
            fn.append(Modifier(sA, sAsys))
        else:
            sAsys = sA
        fn.append(pym.EigenSolve(sAsys, [slambda, sU]))
        sLamsum, sVecsum = pym.Signal("lambdasum"), pym.Signal("vecsum")
        fn.append(pym.EinSum(slambda, sLamsum, expression='i->'))
        fn.append(pym.EinSum(sU, sVecsum, expression='ij->'))
        sSumsum = pym.Signal("allsum")
        fn.append(pym.MathGeneral([sLamsum, sVecsum], sSumsum, expression='inp0 + inp1'))
        fn.response()

        # Test residual and normalization
        self.assertTrue(np.allclose(sAsys.state @ sU.state, sU.state @ np.diag(slambda.state)))
        self.assertTrue(np.allclose(np.diag(sU.state.T @ sU.state), 1.0))

        # Check finite difference (test_fn with quite loose tolerances because of numerical precision)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, sA, [slambda, sU, sLamsum, sVecsum, sSumsum], test_fn=tfn, verbose=True)

    def test_asymmetric_real(self):
        """ Test asymmetric real dense matrix """
        N = 10
        A = np.random.rand(N, N)
        self.use_solver(A)

    def test_asymmetric_complex(self):
        """ Test asymmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        self.use_solver(A)

    # # ------------- Symmetric -------------
    def test_symmetric_real(self):
        """ Test symmetric real dense matrix """
        N = 10
        A = np.random.rand(N, N)
        self.use_solver(A, symmetry=True)

    def test_symmetric_complex(self):
        """ Test symmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        self.use_solver(A, symmetry=True)

    def test_hermitian_complex(self):
        """ Test symmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        self.use_solver(A, hermitian=True)


class TestEigenSolverDense_Generalized(unittest.TestCase):
    def use_solver(self, A, B, symmetry=False, hermitian=False):
        sA = pym.Signal("A", A)
        sB = pym.Signal("B", B)

        slambda = pym.Signal("lambda")
        sU = pym.Signal("eigvec")

        fn = pym.Network()
        if symmetry or hermitian:
            if hermitian and symmetry and np.iscomplexobj(A):
                raise RuntimeError("Complex matrix cannot be symmetric and hermitian at the same time")
            Modifier = Herm if hermitian else Symm
            sAsys = pym.Signal("Asym", A)
            fn.append(Modifier(sA, sAsys))

            sBsys = pym.Signal("Bsym", B)
            fn.append(Modifier(sB, sBsys))
        else:
            sAsys = sA
            sBsys = sB
        fn.append(pym.EigenSolve([sAsys, sBsys], [slambda, sU]))
        sLamsum, sVecsum = pym.Signal("lambdasum"), pym.Signal("vecsum")
        fn.append(pym.EinSum(slambda, sLamsum, expression='i->'))
        fn.append(pym.EinSum(sU, sVecsum, expression='ij->'))
        sSumsum = pym.Signal("allsum")
        fn.append(pym.MathGeneral([sLamsum, sVecsum], sSumsum, expression='inp0 + inp1'))
        fn.response()

        print(f"lambdas = {slambda.state}")

        # Test residual and normalization
        self.assertTrue(np.allclose(sAsys.state @ sU.state, sBsys.state @ sU.state @ np.diag(slambda.state)))
        self.assertTrue(np.allclose(np.diag(sU.state.T @ sBsys.state @ sU.state), 1.0))

        # Check finite difference (test_fn with quite loose tolerances because of numerical precision)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sA, sB], [sLamsum, sVecsum, sSumsum], test_fn=tfn, verbose=True)

    def test_asymmetric_real(self):
        """ Test asymmetric real dense matrix """
        N = 10
        A = np.random.rand(N, N)
        # Matrix B must be positive definite, so we construct it using eigenmodes Q and (positive) eigenvalues D
        Q, _ = np.linalg.qr(np.random.rand(N, N))
        D = np.diag(np.random.rand(N) + 0.3)
        self.use_solver(A, Q@D@Q.T)

    def test_asymmetric_complex(self):
        """ Test asymmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        Q, _ = np.linalg.qr(np.random.rand(N, N) + 1j * np.random.rand(N, N))
        D = np.diag(np.random.rand(N) + 0.3)
        self.use_solver(A, Q@D@Q.T.conj())

    # # ------------- Symmetric -------------
    def test_symmetric_real(self):
        """ Test symmetric real dense matrix """
        N = 10
        A = np.random.rand(N, N)
        Q, _ = np.linalg.qr(np.random.rand(N, N))
        D = np.diag(np.random.rand(N) * 0.9 + 0.1)
        self.use_solver(A, Q @ D @ Q.T, symmetry=True)

    def test_symmetric_complex(self):
        """ Test symmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        Q, _ = np.linalg.qr(np.random.rand(N, N) + 1j * np.random.rand(N, N))
        D = np.diag(np.random.rand(N) * 0.9 + 0.1)
        self.use_solver(A, Q@D@Q.T.conj(), symmetry=True)

    def test_hermitian_complex(self):
        """ Test symmetric complex dense matrix """
        N = 10
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)
        Q, _ = np.linalg.qr(np.random.rand(N, N) + 1j * np.random.rand(N, N))
        D = np.diag(np.random.rand(N) * 0.9 + 0.1)
        self.use_solver(A, Q@D@Q.T.conj(), hermitian=True)


if __name__ == '__main__':
    unittest.main()



import unittest

import pymoto as pym
import numpy as np


class TestEinSum(unittest.TestCase):
    def assert_fd(self, x0, dx, dg_an, dg_fd):
        self.assertAlmostEqual(dg_an, dg_fd, places=6)

    def test_vec_sum(self):
        n = 4

        a_list = [np.random.rand(n), np.random.rand(n)+1j*np.random.rand(n)]
        for a in a_list:
            out_chk = np.sum(a)

            s_a = pym.Signal("vec", a)
            s_out = pym.Signal("sum(a)")

            blk = pym.EinSum(s_a, s_out, expression="i->")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_vec_dot(self):
        n = 4

        a_list = [np.random.rand(n), np.random.rand(n)+1j*np.random.rand(n)]
        b_list = [np.random.rand(n), np.random.rand(n)+1j*np.random.rand(n)]
        for a, b in zip(a_list, b_list):
            out_chk = a@b

            s_a = pym.Signal("veca", a)
            s_b = pym.Signal("vecb", b)
            s_out = pym.Signal("a^T.b")

            blk = pym.EinSum([s_a, s_b], s_out, expression="i,i->")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_vec_outer(self):
        n = 4

        a_list = [np.random.rand(n), np.random.rand(n)+1j*np.random.rand(n)]
        b_list = [np.random.rand(n), np.random.rand(n)+1j*np.random.rand(n)]
        for a, b in zip(a_list, b_list):
            out_chk = np.outer(a, b)

            s_a = pym.Signal("veca", a)
            s_b = pym.Signal("vecb", b)
            s_out = pym.Signal("a.b^T")

            blk = pym.EinSum([s_a, s_b], s_out, expression="i,j->ij")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_mat_trace(self):
        n = 4

        a_list = [np.random.rand(n, n), np.random.rand(n, n)+1j*np.random.rand(n, n)]

        for a in a_list:
            out_chk = np.trace(a)

            s_a = pym.Signal("mat", a)
            s_out = pym.Signal("trace(A)")

            blk = pym.EinSum(s_a, s_out, expression="ii->")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_mat_sum(self):
        n = 3

        a_list = [np.random.rand(n, n), np.random.rand(n, n)+1j*np.random.rand(n, n)]
        for a in a_list:
            out_chk = np.sum(a)

            s_a = pym.Signal("mat", a)
            s_out = pym.Signal("sum(A)")

            blk = pym.EinSum(s_a, s_out, expression="ij->")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    # TODO: Sensitivity does not work for this
    # def test_einsum_mat_diag(self):
    #     n = 3
    #
    #     a = np.random.rand(n, n)
    #
    #     out_chk = np.diag(a)
    #
    #     s_a = pym.Signal("mat", a)
    #     s_out = pym.Signal("diag(A)")
    #
    #     blk = pym.EinSum(s_a, s_out, expression="ii->i")
    #
    #     blk.response()
    #     self.assertTrue(np.allclose(out_chk, s_out.state))
    #     pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_matmat(self):
        n = 3

        a_list = [np.random.rand(n, n), np.random.rand(n, n)+1j*np.random.rand(n, n)]
        b_list = [np.random.rand(n, n), np.random.rand(n, n)+1j*np.random.rand(n, n)]
        for a, b in zip(a_list, b_list):
            out_chk = a.dot(b)

            s_a = pym.Signal("matA", a)
            s_b = pym.Signal("matB", b)
            s_out = pym.Signal("matC")

            blk = pym.EinSum([s_a, s_b], s_out, expression="ij,jk->ik")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)

    def test_matmatvec(self):
        N = 5
        p = 2

        A_list = [np.random.rand(N, p), 1j*np.random.rand(N, p), np.random.rand(N, p) + 1j*np.random.rand(N, p)]
        B_list = [np.random.rand(p, p), 1j*np.random.rand(p, p), np.random.rand(p, p) + 1j*np.random.rand(p, p)]
        v_list = [np.random.rand(p), 1j*np.random.rand(p), np.random.rand(p)+1j*np.random.rand(p)]

        for A, B, v in zip(A_list, B_list, v_list):
            out_chk = (A.dot(B)).dot(np.diag(v))

            s_A = pym.Signal("A", A)
            s_B = pym.Signal("B", B)
            s_v = pym.Signal("v", v)
            s_out = pym.Signal("C")

            blk = pym.EinSum([s_A, s_B, s_v], s_out, expression="ij,jk,k->ik")

            blk.response()
            self.assertTrue(np.allclose(out_chk, s_out.state))
            pym.finite_difference(blk, test_fn=self.assert_fd)


if __name__ == '__main__':
    unittest.main()



from typing import Any
import numpy as np


def _parse_to_list(*args: Any):
    if len(args) == 0:
        return []
    elif len(args) == 1:
        var_in = args[0]
    else:
        var_in = args

    if var_in is None:
        return []
    elif isinstance(var_in, list):
        return var_in
    elif isinstance(var_in, tuple) or isinstance(var_in, set):
        return list(var_in)
    else:
        return [var_in]


def _concatenate_to_array(var_list: list):
    values = np.array([])
    cumulative_inds = np.zeros(len(var_list)+1, dtype=int)

    for i, v in enumerate(var_list):
        if v is None:
            raise ValueError("Trying to add None to the array")

        values = np.append(values, v)
        cumulative_inds[i+1] = len(values)

    return values, cumulative_inds



from abc import ABC

import numpy as np
from pymoto import Module
from ..utils import _parse_to_list

try:  # AutoDiff module
    import jax
    _has_jax = True
except ImportError as e:
    _jax_error = e
    _has_jax = False


class AutoMod(Module, ABC):
    """ Module that automatically differentiates the response function """

    def __init__(self, *args, **kwargs):
        if not _has_jax:
            raise ImportError(f"Could not create this object, as dependency \"jax\" cannot be found on this system. "
                              f"Import failed with error: {_jax_error}")
        super().__init__(*args, **kwargs)

    def response(self):
        # Calculate the response and tangent operator (JAX Vector-Jacobian product)
        y, self.vjp_fn = jax.vjp(self._response, *[s.state for s in self.sig_in])
        y = _parse_to_list(y)

        # Assign all the states
        for i, s in enumerate(self.sig_out):
            s.state = y[i]

    def sensitivity(self):
        # Gather the output sensitivities
        dfdv = [s.sensitivity for s in self.sig_out]
        if np.all([df is None for df in dfdv]):
            return
        for i in range(len(dfdv)):
            if dfdv[i] is None:  # JAX does not accept None as 0
                dfdv[i] = np.zeros_like(self.sig_out[i].state)

        dfdv = tuple(dfdv) if len(dfdv) > 1 else dfdv[0]

        # Calculate backward sensitivity
        dfdx = _parse_to_list(self.vjp_fn(dfdv))

        # Assign the sensitivities
        for i, s in enumerate(self.sig_in):
            if np.iscomplexobj(dfdx[i]):
                s.add_sensitivity(dfdx[i])
            else:
                s.add_sensitivity(dfdx[i])
